#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock
from unittest.mock import PropertyMock

import pytest
from google.api_core.retry import Retry

from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.hooks.compute import ComputeEngineHook, GceOperationStatus

from unit.google.cloud.utils.base_gcp_mock import (
    GCP_PROJECT_ID_HOOK_UNIT_TEST,
    mock_base_gcp_hook_default_project_id,
    mock_base_gcp_hook_no_default_project_id,
)

GCE_ZONE = "zone"
GCE_INSTANCE = "instance"
GCE_INSTANCE_TEMPLATE = "instance-template"
GCE_REQUEST_ID = "request_id"
GCE_INSTANCE_GROUP_MANAGER = "instance_group_manager"

PROJECT_ID = "project-id"
RESOURCE_ID = "resource-id"
BODY = {"body": "test"}
ZONE = "zone"
SOURCE_INSTANCE_TEMPLATE = "source-instance-template"
TIMEOUT = 120
RETRY = mock.MagicMock(Retry)
METADATA = [("key", "value")]

GCP_CONN_ID = "test-conn"
IMPERSONATION_CHAIN = ["ACCOUNT_1", "ACCOUNT_2", "ACCOUNT_3@google.com"]
API_VERSION = "v1"

BASE_STRING = "airflow.providers.google.common.hooks.base_google.{}"
COMPUTE_ENGINE_HOOK_PATH = "airflow.providers.google.cloud.hooks.compute.{}"


class TestGcpComputeHookApiCall:
    def setup_method(self):
        with mock.patch(
            BASE_STRING.format("GoogleBaseHook.__init__"),
            new=mock_base_gcp_hook_default_project_id,
        ):
            self.hook = ComputeEngineHook(
                gcp_conn_id=GCP_CONN_ID,
                api_version=API_VERSION,
                impersonation_chain=IMPERSONATION_CHAIN,
            )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH.format("ComputeEngineHook._wait_for_operation_to_complete"))
    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value="mocked-google",
    )
    @mock.patch(COMPUTE_ENGINE_HOOK_PATH.format("ComputeEngineHook.get_compute_instance_template_client"))
    def test_insert_template_should_execute_successfully(
        self, mock_client, mocked_project_id, wait_for_operation_to_complete
    ):
        wait_for_operation_to_complete.return_value = None
        self.hook.insert_instance_template(
            project_id=PROJECT_ID,
            body=BODY,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )
        mock_client.return_value.insert.assert_called_once_with(
            request=dict(
                instance_template_resource=BODY,
                project=PROJECT_ID,
                request_id=None,
            ),
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH.format("ComputeEngineHook._wait_for_operation_to_complete"))
    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value="mocked-google",
    )
    @mock.patch(COMPUTE_ENGINE_HOOK_PATH.format("ComputeEngineHook.get_compute_instance_template_client"))
    def test_insert_template_should_not_throw_ex_when_project_id_none(
        self, mock_client, mocked_project_id, wait_for_operation_to_complete
    ):
        wait_for_operation_to_complete.return_value = None
        self.hook.insert_instance_template(
            body=BODY,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )
        mock_client.return_value.insert.assert_called_once_with(
            request=dict(
                instance_template_resource=BODY,
                request_id=None,
                project="mocked-google",
            ),
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH.format("ComputeEngineHook._wait_for_operation_to_complete"))
    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value="mocked-google",
    )
    @mock.patch(COMPUTE_ENGINE_HOOK_PATH.format("ComputeEngineHook.get_compute_instance_template_client"))
    def test_delete_template_should_execute_successfully(
        self, mock_client, mocked_project_id, wait_for_operation_to_complete
    ):
        wait_for_operation_to_complete.return_value = None
        self.hook.delete_instance_template(
            project_id=PROJECT_ID,
            resource_id=RESOURCE_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )
        mock_client.return_value.delete.assert_called_once_with(
            request=dict(
                instance_template=RESOURCE_ID,
                project=PROJECT_ID,
                request_id=None,
            ),
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH.format("ComputeEngineHook._wait_for_operation_to_complete"))
    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value="mocked-google",
    )
    @mock.patch(COMPUTE_ENGINE_HOOK_PATH.format("ComputeEngineHook.get_compute_instance_template_client"))
    def test_delete_template_should_not_throw_ex_when_project_id_none(
        self, mock_client, mocked_project_id, wait_for_operation_to_complete
    ):
        wait_for_operation_to_complete.return_value = None
        self.hook.delete_instance_template(
            resource_id=RESOURCE_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )
        mock_client.return_value.delete.assert_called_once_with(
            request=dict(
                instance_template=RESOURCE_ID,
                project="mocked-google",
                request_id=None,
            ),
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH.format("ComputeEngineHook.get_compute_instance_template_client"))
    def test_get_template_should_execute_successfully(self, mock_client):
        self.hook.get_instance_template(
            project_id=PROJECT_ID,
            resource_id=RESOURCE_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )
        mock_client.return_value.get.assert_called_once_with(
            request=dict(
                instance_template=RESOURCE_ID,
                project=PROJECT_ID,
            ),
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value="mocked-google",
    )
    @mock.patch(COMPUTE_ENGINE_HOOK_PATH.format("ComputeEngineHook.get_compute_instance_template_client"))
    def test_get_template_should_not_throw_ex_when_project_id_none(self, mock_client, mocked_project_id):
        self.hook.get_instance_template(
            resource_id=RESOURCE_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )
        mock_client.return_value.get.assert_called_once_with(
            request=dict(
                instance_template=RESOURCE_ID,
                project="mocked-google",
            ),
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH.format("ComputeEngineHook._wait_for_operation_to_complete"))
    @mock.patch(COMPUTE_ENGINE_HOOK_PATH.format("ComputeEngineHook.get_compute_instance_client"))
    def test_insert_instance_should_execute_successfully(self, mock_client, wait_for_operation_to_complete):
        wait_for_operation_to_complete.return_value = None
        self.hook.insert_instance(
            project_id=PROJECT_ID,
            body=BODY,
            zone=ZONE,
            source_instance_template=SOURCE_INSTANCE_TEMPLATE,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )
        mock_client.return_value.insert.assert_called_once_with(
            request=dict(
                instance_resource=BODY,
                project=PROJECT_ID,
                request_id=None,
                zone=ZONE,
                source_instance_template=SOURCE_INSTANCE_TEMPLATE,
            ),
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH.format("ComputeEngineHook._wait_for_operation_to_complete"))
    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value="mocked-google",
    )
    @mock.patch(COMPUTE_ENGINE_HOOK_PATH.format("ComputeEngineHook.get_compute_instance_client"))
    def test_insert_instance_should_not_throw_ex_when_project_id_none(
        self, mock_client, mocked_project_id, wait_for_operation_to_complete
    ):
        wait_for_operation_to_complete.return_value = None
        self.hook.insert_instance(
            body=BODY,
            zone=ZONE,
            source_instance_template=SOURCE_INSTANCE_TEMPLATE,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )
        mock_client.return_value.insert.assert_called_once_with(
            request=dict(
                instance_resource=BODY,
                project="mocked-google",
                request_id=None,
                zone=ZONE,
                source_instance_template=SOURCE_INSTANCE_TEMPLATE,
            ),
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH.format("ComputeEngineHook.get_compute_instance_client"))
    def test_get_instance_should_execute_successfully(self, mock_client):
        self.hook.get_instance(
            resource_id=RESOURCE_ID,
            zone=ZONE,
            project_id=PROJECT_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )
        mock_client.return_value.get.assert_called_once_with(
            request=dict(
                instance=RESOURCE_ID,
                project=PROJECT_ID,
                zone=ZONE,
            ),
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value="mocked-google",
    )
    @mock.patch(COMPUTE_ENGINE_HOOK_PATH.format("ComputeEngineHook.get_compute_instance_client"))
    def test_get_instance_should_not_throw_ex_when_project_id_none(self, mock_client, mocked_project_id):
        self.hook.get_instance(
            resource_id=RESOURCE_ID,
            zone=ZONE,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )
        mock_client.return_value.get.assert_called_once_with(
            request=dict(
                instance=RESOURCE_ID,
                project="mocked-google",
                zone=ZONE,
            ),
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH.format("ComputeEngineHook._wait_for_operation_to_complete"))
    @mock.patch(COMPUTE_ENGINE_HOOK_PATH.format("ComputeEngineHook.get_compute_instance_client"))
    def test_delete_instance_should_execute_successfully(self, mock_client, wait_for_operation_to_complete):
        wait_for_operation_to_complete.return_value = None
        self.hook.delete_instance(
            resource_id=RESOURCE_ID,
            zone=ZONE,
            project_id=PROJECT_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )
        mock_client.return_value.delete.assert_called_once_with(
            request=dict(
                instance=RESOURCE_ID,
                project=PROJECT_ID,
                request_id=None,
                zone=ZONE,
            ),
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH.format("ComputeEngineHook._wait_for_operation_to_complete"))
    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value="mocked-google",
    )
    @mock.patch(COMPUTE_ENGINE_HOOK_PATH.format("ComputeEngineHook.get_compute_instance_client"))
    def test_delete_instance_should_not_throw_ex_when_project_id_none(
        self, mock_client, mocked_project_id, wait_for_operation_to_complete
    ):
        wait_for_operation_to_complete.return_value = None
        self.hook.delete_instance(
            resource_id=RESOURCE_ID,
            zone=ZONE,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )
        mock_client.return_value.delete.assert_called_once_with(
            request=dict(
                instance=RESOURCE_ID,
                project="mocked-google",
                request_id=None,
                zone=ZONE,
            ),
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH.format("ComputeEngineHook._wait_for_operation_to_complete"))
    @mock.patch(
        COMPUTE_ENGINE_HOOK_PATH.format("ComputeEngineHook.get_compute_instance_group_managers_client")
    )
    def test_insert_instance_group_manager_should_execute_successfully(
        self, mock_client, wait_for_operation_to_complete
    ):
        wait_for_operation_to_complete.return_value = None
        self.hook.insert_instance_group_manager(
            body=BODY,
            zone=ZONE,
            project_id=PROJECT_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )
        mock_client.return_value.insert.assert_called_once_with(
            request=dict(
                instance_group_manager_resource=BODY,
                project=PROJECT_ID,
                request_id=None,
                zone=ZONE,
            ),
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH.format("ComputeEngineHook._wait_for_operation_to_complete"))
    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value="mocked-google",
    )
    @mock.patch(
        COMPUTE_ENGINE_HOOK_PATH.format("ComputeEngineHook.get_compute_instance_group_managers_client")
    )
    def test_insert_instance_group_manager_should_not_throw_ex_when_project_id_none(
        self, mock_client, mocked_project_id, wait_for_operation_to_complete
    ):
        wait_for_operation_to_complete.return_value = None
        self.hook.insert_instance_group_manager(
            body=BODY,
            zone=ZONE,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )
        mock_client.return_value.insert.assert_called_once_with(
            request=dict(
                instance_group_manager_resource=BODY,
                project="mocked-google",
                request_id=None,
                zone=ZONE,
            ),
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

    @mock.patch(
        COMPUTE_ENGINE_HOOK_PATH.format("ComputeEngineHook.get_compute_instance_group_managers_client")
    )
    def test_get_instance_group_manager_should_execute_successfully(self, mock_client):
        self.hook.get_instance_group_manager(
            resource_id=RESOURCE_ID,
            zone=ZONE,
            project_id=PROJECT_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )
        mock_client.return_value.get.assert_called_once_with(
            request=dict(
                instance_group_manager=RESOURCE_ID,
                project=PROJECT_ID,
                zone=ZONE,
            ),
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value="mocked-google",
    )
    @mock.patch(
        COMPUTE_ENGINE_HOOK_PATH.format("ComputeEngineHook.get_compute_instance_group_managers_client")
    )
    def test_get_instance_group_manager_should_not_throw_ex_when_project_id_none(
        self, mock_client, mocked_project_id
    ):
        self.hook.get_instance_group_manager(
            resource_id=RESOURCE_ID,
            zone=ZONE,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )
        mock_client.return_value.get.assert_called_once_with(
            request=dict(
                instance_group_manager=RESOURCE_ID,
                project="mocked-google",
                zone=ZONE,
            ),
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH.format("ComputeEngineHook._wait_for_operation_to_complete"))
    @mock.patch(
        COMPUTE_ENGINE_HOOK_PATH.format("ComputeEngineHook.get_compute_instance_group_managers_client")
    )
    def test_delete_instance_group_manager_should_execute_successfully(
        self, mock_client, wait_for_operation_to_complete
    ):
        wait_for_operation_to_complete.return_value = None
        self.hook.delete_instance_group_manager(
            resource_id=RESOURCE_ID,
            zone=ZONE,
            project_id=PROJECT_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )
        mock_client.return_value.delete.assert_called_once_with(
            request=dict(
                instance_group_manager=RESOURCE_ID,
                project=PROJECT_ID,
                zone=ZONE,
                request_id=None,
            ),
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH.format("ComputeEngineHook._wait_for_operation_to_complete"))
    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value="mocked-google",
    )
    @mock.patch(
        COMPUTE_ENGINE_HOOK_PATH.format("ComputeEngineHook.get_compute_instance_group_managers_client")
    )
    def test_delete_instance_group_manager_should_not_throw_ex_when_project_id_none(
        self, mock_client, mocked_project_id, wait_for_operation_to_complete
    ):
        wait_for_operation_to_complete.return_value = None
        self.hook.delete_instance_group_manager(
            resource_id=RESOURCE_ID,
            zone=ZONE,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )
        mock_client.return_value.delete.assert_called_once_with(
            request=dict(
                instance_group_manager=RESOURCE_ID,
                project="mocked-google",
                zone=ZONE,
                request_id=None,
            ),
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )


class TestGcpComputeHookNoDefaultProjectId:
    def setup_method(self):
        with mock.patch(
            "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
            new=mock_base_gcp_hook_no_default_project_id,
        ):
            self.gce_hook_no_project_id = ComputeEngineHook(gcp_conn_id="test")

    @mock.patch("airflow.providers.google.cloud.hooks.compute.ComputeEngineHook._authorize")
    @mock.patch("airflow.providers.google.cloud.hooks.compute.build")
    def test_gce_client_creation(self, mock_build, mock_authorize):
        result = self.gce_hook_no_project_id.get_conn()
        mock_build.assert_called_once_with(
            "compute", "v1", http=mock_authorize.return_value, cache_discovery=False
        )
        assert mock_build.return_value == result

    @mock.patch("airflow.providers.google.cloud.hooks.compute.ComputeEngineHook.get_conn")
    @mock.patch(
        "airflow.providers.google.cloud.hooks.compute.ComputeEngineHook._wait_for_operation_to_complete"
    )
    def test_start_instance_overridden_project_id(self, wait_for_operation_to_complete, get_conn):
        start_method = get_conn.return_value.instances.return_value.start
        execute_method = start_method.return_value.execute
        execute_method.return_value = {"name": "operation_id"}
        wait_for_operation_to_complete.return_value = None
        res = self.gce_hook_no_project_id.start_instance(
            project_id="example-project", zone=GCE_ZONE, resource_id=GCE_INSTANCE
        )
        assert res is None
        start_method.assert_called_once_with(instance="instance", project="example-project", zone="zone")
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_called_once_with(
            project_id="example-project", operation_name="operation_id", zone="zone"
        )

    @mock.patch("airflow.providers.google.cloud.hooks.compute.ComputeEngineHook.get_conn")
    @mock.patch(
        "airflow.providers.google.cloud.hooks.compute.ComputeEngineHook._wait_for_operation_to_complete"
    )
    def test_stop_instance_overridden_project_id(self, wait_for_operation_to_complete, get_conn):
        stop_method = get_conn.return_value.instances.return_value.stop
        execute_method = stop_method.return_value.execute
        execute_method.return_value = {"name": "operation_id"}
        wait_for_operation_to_complete.return_value = None
        res = self.gce_hook_no_project_id.stop_instance(
            project_id="example-project", zone=GCE_ZONE, resource_id=GCE_INSTANCE
        )
        assert res is None
        stop_method.assert_called_once_with(instance="instance", project="example-project", zone="zone")
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_called_once_with(
            project_id="example-project", operation_name="operation_id", zone="zone"
        )

    @mock.patch("airflow.providers.google.cloud.hooks.compute.ComputeEngineHook.get_conn")
    @mock.patch(
        "airflow.providers.google.cloud.hooks.compute.ComputeEngineHook._wait_for_operation_to_complete"
    )
    def test_set_machine_type_overridden_project_id(self, wait_for_operation_to_complete, get_conn):
        set_machine_type_method = get_conn.return_value.instances.return_value.setMachineType
        execute_method = set_machine_type_method.return_value.execute
        execute_method.return_value = {"name": "operation_id"}
        wait_for_operation_to_complete.return_value = None
        res = self.gce_hook_no_project_id.set_machine_type(
            body={}, project_id="example-project", zone=GCE_ZONE, resource_id=GCE_INSTANCE
        )
        assert res is None
        set_machine_type_method.assert_called_once_with(
            body={}, instance="instance", project="example-project", zone="zone"
        )
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_called_once_with(
            project_id="example-project", operation_name="operation_id", zone="zone"
        )

    @mock.patch("airflow.providers.google.cloud.hooks.compute.ComputeEngineHook.get_conn")
    @mock.patch(
        "airflow.providers.google.cloud.hooks.compute.ComputeEngineHook._wait_for_operation_to_complete"
    )
    def test_patch_instance_group_manager_overridden_project_id(
        self, wait_for_operation_to_complete, get_conn
    ):
        patch_method = get_conn.return_value.instanceGroupManagers.return_value.patch
        execute_method = patch_method.return_value.execute
        execute_method.return_value = {"name": "operation_id"}
        wait_for_operation_to_complete.return_value = None
        res = self.gce_hook_no_project_id.patch_instance_group_manager(
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
            zone=GCE_ZONE,
            resource_id=GCE_INSTANCE_GROUP_MANAGER,
            body={},
            request_id=GCE_REQUEST_ID,
        )
        assert res is None
        patch_method.assert_called_once_with(
            body={},
            instanceGroupManager="instance_group_manager",
            project="example-project",
            requestId="request_id",
            zone="zone",
        )
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_called_once_with(
            operation_name="operation_id", project_id="example-project", zone="zone"
        )


class TestGcpComputeHookDefaultProjectId:
    def setup_method(self):
        with mock.patch(
            "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
            new=mock_base_gcp_hook_default_project_id,
        ):
            self.gce_hook = ComputeEngineHook(gcp_conn_id="test")

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.compute.ComputeEngineHook.get_conn")
    @mock.patch(
        "airflow.providers.google.cloud.hooks.compute.ComputeEngineHook._wait_for_operation_to_complete"
    )
    def test_start_instance(self, wait_for_operation_to_complete, get_conn, mock_project_id):
        start_method = get_conn.return_value.instances.return_value.start
        execute_method = start_method.return_value.execute
        execute_method.return_value = {"name": "operation_id"}
        wait_for_operation_to_complete.return_value = None
        res = self.gce_hook.start_instance(
            zone=GCE_ZONE,
            resource_id=GCE_INSTANCE,
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
        )
        assert res is None
        start_method.assert_called_once_with(instance="instance", project="example-project", zone="zone")
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_called_once_with(
            project_id="example-project", operation_name="operation_id", zone="zone"
        )

    @mock.patch("airflow.providers.google.cloud.hooks.compute.ComputeEngineHook.get_conn")
    @mock.patch(
        "airflow.providers.google.cloud.hooks.compute.ComputeEngineHook._wait_for_operation_to_complete"
    )
    def test_start_instance_overridden_project_id(self, wait_for_operation_to_complete, get_conn):
        start_method = get_conn.return_value.instances.return_value.start
        execute_method = start_method.return_value.execute
        execute_method.return_value = {"name": "operation_id"}
        wait_for_operation_to_complete.return_value = None
        res = self.gce_hook.start_instance(project_id="new-project", zone=GCE_ZONE, resource_id=GCE_INSTANCE)
        assert res is None
        start_method.assert_called_once_with(instance="instance", project="new-project", zone="zone")
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_called_once_with(
            project_id="new-project", operation_name="operation_id", zone="zone"
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.compute.ComputeEngineHook.get_conn")
    @mock.patch(
        "airflow.providers.google.cloud.hooks.compute.ComputeEngineHook._wait_for_operation_to_complete"
    )
    def test_stop_instance(self, wait_for_operation_to_complete, get_conn, mock_project_id):
        stop_method = get_conn.return_value.instances.return_value.stop
        execute_method = stop_method.return_value.execute
        execute_method.return_value = {"name": "operation_id"}
        wait_for_operation_to_complete.return_value = None
        res = self.gce_hook.stop_instance(
            zone=GCE_ZONE,
            resource_id=GCE_INSTANCE,
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
        )
        assert res is None
        stop_method.assert_called_once_with(instance="instance", project="example-project", zone="zone")
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_called_once_with(
            project_id="example-project", operation_name="operation_id", zone="zone"
        )

    @mock.patch("airflow.providers.google.cloud.hooks.compute.ComputeEngineHook.get_conn")
    @mock.patch(
        "airflow.providers.google.cloud.hooks.compute.ComputeEngineHook._wait_for_operation_to_complete"
    )
    def test_stop_instance_overridden_project_id(self, wait_for_operation_to_complete, get_conn):
        stop_method = get_conn.return_value.instances.return_value.stop
        execute_method = stop_method.return_value.execute
        execute_method.return_value = {"name": "operation_id"}
        wait_for_operation_to_complete.return_value = None
        res = self.gce_hook.stop_instance(project_id="new-project", zone=GCE_ZONE, resource_id=GCE_INSTANCE)
        assert res is None
        stop_method.assert_called_once_with(instance="instance", project="new-project", zone="zone")
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_called_once_with(
            project_id="new-project", operation_name="operation_id", zone="zone"
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.compute.ComputeEngineHook.get_conn")
    @mock.patch(
        "airflow.providers.google.cloud.hooks.compute.ComputeEngineHook._wait_for_operation_to_complete"
    )
    def test_set_machine_type_instance(self, wait_for_operation_to_complete, get_conn, mock_project_id):
        execute_method = get_conn.return_value.instances.return_value.setMachineType.return_value.execute
        execute_method.return_value = {"name": "operation_id"}
        wait_for_operation_to_complete.return_value = None
        res = self.gce_hook.set_machine_type(
            body={},
            zone=GCE_ZONE,
            resource_id=GCE_INSTANCE,
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
        )
        assert res is None
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_called_once_with(
            project_id="example-project", operation_name="operation_id", zone="zone"
        )

    @mock.patch("airflow.providers.google.cloud.hooks.compute.ComputeEngineHook.get_conn")
    @mock.patch(
        "airflow.providers.google.cloud.hooks.compute.ComputeEngineHook._wait_for_operation_to_complete"
    )
    def test_set_machine_type_instance_overridden_project_id(self, wait_for_operation_to_complete, get_conn):
        execute_method = get_conn.return_value.instances.return_value.setMachineType.return_value.execute
        execute_method.return_value = {"name": "operation_id"}
        wait_for_operation_to_complete.return_value = None
        res = self.gce_hook.set_machine_type(
            project_id="new-project", body={}, zone=GCE_ZONE, resource_id=GCE_INSTANCE
        )
        assert res is None
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_called_once_with(
            project_id="new-project", operation_name="operation_id", zone="zone"
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.compute.ComputeEngineHook.get_conn")
    @mock.patch(
        "airflow.providers.google.cloud.hooks.compute.ComputeEngineHook._wait_for_operation_to_complete"
    )
    def test_patch_instance_group_manager(self, wait_for_operation_to_complete, get_conn, mock_project_id):
        patch_method = get_conn.return_value.instanceGroupManagers.return_value.patch
        execute_method = patch_method.return_value.execute
        execute_method.return_value = {"name": "operation_id"}
        wait_for_operation_to_complete.return_value = None
        res = self.gce_hook.patch_instance_group_manager(
            zone=GCE_ZONE,
            resource_id=GCE_INSTANCE_GROUP_MANAGER,
            body={},
            request_id=GCE_REQUEST_ID,
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
        )
        assert res is None
        patch_method.assert_called_once_with(
            body={},
            instanceGroupManager="instance_group_manager",
            project="example-project",
            requestId="request_id",
            zone="zone",
        )
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_called_once_with(
            operation_name="operation_id", project_id="example-project", zone="zone"
        )

    @mock.patch("airflow.providers.google.cloud.hooks.compute.ComputeEngineHook.get_conn")
    @mock.patch(
        "airflow.providers.google.cloud.hooks.compute.ComputeEngineHook._wait_for_operation_to_complete"
    )
    def test_patch_instance_group_manager_overridden_project_id(
        self, wait_for_operation_to_complete, get_conn
    ):
        patch_method = get_conn.return_value.instanceGroupManagers.return_value.patch
        execute_method = patch_method.return_value.execute
        execute_method.return_value = {"name": "operation_id"}
        wait_for_operation_to_complete.return_value = None
        res = self.gce_hook.patch_instance_group_manager(
            project_id="new-project",
            zone=GCE_ZONE,
            resource_id=GCE_INSTANCE_GROUP_MANAGER,
            body={},
            request_id=GCE_REQUEST_ID,
        )
        assert res is None
        patch_method.assert_called_once_with(
            body={},
            instanceGroupManager="instance_group_manager",
            project="new-project",
            requestId="request_id",
            zone="zone",
        )
        execute_method.assert_called_once_with(num_retries=5)
        wait_for_operation_to_complete.assert_called_once_with(
            operation_name="operation_id", project_id="new-project", zone="zone"
        )

    @mock.patch("airflow.providers.google.cloud.hooks.compute.ComputeEngineHook.get_conn")
    @mock.patch(
        "airflow.providers.google.cloud.hooks.compute.ComputeEngineHook._check_global_operation_status"
    )
    def test_wait_for_operation_to_complete_no_zone(self, mock_operation_status, mock_get_conn):
        service = "test-service"
        project_id = "test-project"
        operation_name = "test-operation"
        num_retries = self.gce_hook.num_retries

        # Test success
        mock_get_conn.return_value = service
        mock_operation_status.return_value = {"status": GceOperationStatus.DONE, "error": None}
        self.gce_hook._wait_for_operation_to_complete(
            project_id=project_id, operation_name=operation_name, zone=None
        )

        mock_operation_status.assert_called_once_with(
            service=service, operation_name=operation_name, project_id=project_id, num_retries=num_retries
        )

    @mock.patch("airflow.providers.google.cloud.hooks.compute.ComputeEngineHook.get_conn")
    @mock.patch(
        "airflow.providers.google.cloud.hooks.compute.ComputeEngineHook._check_global_operation_status"
    )
    def test_wait_for_operation_to_complete_no_zone_error(self, mock_operation_status, mock_get_conn):
        service = "test-service"
        project_id = "test-project"
        operation_name = "test-operation"

        # Test error
        mock_get_conn.return_value = service
        mock_operation_status.return_value = {
            "status": GceOperationStatus.DONE,
            "error": {"errors": "some nasty errors"},
            "httpErrorStatusCode": 400,
            "httpErrorMessage": "sample msg",
        }

        with pytest.raises(AirflowException):
            self.gce_hook._wait_for_operation_to_complete(
                project_id=project_id, operation_name=operation_name, zone=None
            )

    @mock.patch("airflow.providers.google.cloud.hooks.compute.ComputeEngineHook.get_conn")
    @mock.patch("airflow.providers.google.cloud.hooks.compute.ComputeEngineHook._check_zone_operation_status")
    def test_wait_for_operation_to_complete_with_zone(self, mock_operation_status, mock_get_conn):
        service = "test-service"
        project_id = "test-project"
        operation_name = "test-operation"
        zone = "west-europe3"
        num_retries = self.gce_hook.num_retries

        # Test success
        mock_get_conn.return_value = service
        mock_operation_status.return_value = {"status": GceOperationStatus.DONE, "error": None}
        self.gce_hook._wait_for_operation_to_complete(
            project_id=project_id, operation_name=operation_name, zone=zone
        )

        mock_operation_status.assert_called_once_with(service, operation_name, project_id, zone, num_retries)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
from json import JSONDecodeError

from wtforms.validators import ValidationError


class ValidJson:
    """
    Validates data is valid JSON.

    :param message:
        Error message to raise in case of a validation error.
    """

    def __init__(self, message=None):
        self.message = message

    def __call__(self, form, field):
        if field.data:
            try:
                json.loads(field.data)
            except JSONDecodeError as ex:
                message = self.message or f"JSON Validation Error: {ex}"
                raise ValidationError(message=field.gettext(message.format(field.data)))



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example Airflow DAG that shows how to use FacebookAdsReportToGcsOperator.
"""

from __future__ import annotations

import os
from datetime import datetime

from facebook_business.adobjects.adsinsights import AdsInsights

from airflow.models.baseoperator import chain
from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.bigquery import (
    BigQueryCreateEmptyDatasetOperator,
    BigQueryCreateTableOperator,
    BigQueryDeleteDatasetOperator,
    BigQueryInsertJobOperator,
)
from airflow.providers.google.cloud.operators.gcs import GCSCreateBucketOperator, GCSDeleteBucketOperator
from airflow.providers.google.cloud.transfers.facebook_ads_to_gcs import FacebookAdsReportToGcsOperator
from airflow.providers.google.cloud.transfers.gcs_to_bigquery import GCSToBigQueryOperator

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.google import DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT") or DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID

DAG_ID = "facebook_ads_to_gcs"

DATASET_NAME = f"dataset_{DAG_ID}_{ENV_ID}"
BUCKET_NAME = f"bucket_{DAG_ID}_{ENV_ID}"
TABLE_NAME = "airflow_test_datatable"
GCS_OBJ_PATH = "Temp/this_is_my_report_csv.csv"

FIELDS = [
    AdsInsights.Field.campaign_name,
    AdsInsights.Field.campaign_id,
    AdsInsights.Field.ad_id,
    AdsInsights.Field.clicks,
    AdsInsights.Field.impressions,
]
PARAMETERS = {"level": "ad", "date_preset": "yesterday"}

with DAG(
    DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example", "facebook_ads_to_gcs"],
) as dag:
    create_bucket = GCSCreateBucketOperator(
        task_id="create_bucket",
        bucket_name=BUCKET_NAME,
        project_id=PROJECT_ID,
    )

    create_dataset = BigQueryCreateEmptyDatasetOperator(
        task_id="create_dataset",
        dataset_id=DATASET_NAME,
    )

    create_table = BigQueryCreateTableOperator(
        task_id="create_table",
        dataset_id=DATASET_NAME,
        table_id=TABLE_NAME,
        table_resource={
            "schema": {
                "fields": [
                    {"name": "campaign_name", "type": "STRING", "mode": "NULLABLE"},
                    {"name": "campaign_id", "type": "STRING", "mode": "NULLABLE"},
                    {"name": "ad_id", "type": "STRING", "mode": "NULLABLE"},
                    {"name": "clicks", "type": "STRING", "mode": "NULLABLE"},
                    {"name": "impressions", "type": "STRING", "mode": "NULLABLE"},
                ],
            },
        },
    )

    # [START howto_operator_facebook_ads_to_gcs]
    run_operator = FacebookAdsReportToGcsOperator(
        task_id="run_fetch_data",
        owner="airflow",
        bucket_name=BUCKET_NAME,
        parameters=PARAMETERS,
        fields=FIELDS,
        object_name=GCS_OBJ_PATH,
    )
    # [END howto_operator_facebook_ads_to_gcs]

    load_csv = GCSToBigQueryOperator(
        task_id="gcs_to_bq_example",
        bucket=BUCKET_NAME,
        source_objects=[GCS_OBJ_PATH],
        destination_project_dataset_table=f"{DATASET_NAME}.{TABLE_NAME}",
        write_disposition="WRITE_TRUNCATE",
    )

    read_data_from_gcs_many_chunks = BigQueryInsertJobOperator(
        task_id="read_data_from_gcs_many_chunks",
        configuration={
            "query": {
                "query": f"SELECT COUNT(*) FROM `{PROJECT_ID}.{DATASET_NAME}.{TABLE_NAME}`",
                "useLegacySql": False,
            }
        },
    )

    delete_bucket = GCSDeleteBucketOperator(
        task_id="delete_bucket",
        bucket_name=BUCKET_NAME,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    delete_dataset = BigQueryDeleteDatasetOperator(
        task_id="delete_dataset",
        project_id=PROJECT_ID,
        dataset_id=DATASET_NAME,
        delete_contents=True,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    chain(
        # TEST SETUP
        create_bucket,
        create_dataset,
        create_table,
        # TEST BODY
        run_operator,
        load_csv,
        read_data_from_gcs_many_chunks,
        # TEST TEARDOWN
        delete_bucket,
        delete_dataset,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest.mock import MagicMock, PropertyMock, patch

import pytest

from airflow.providers.google.cloud.hooks.secret_manager import GoogleCloudSecretManagerHook
from airflow.providers.google.common.consts import CLIENT_INFO

from unit.google.cloud.utils.base_gcp_mock import GCP_PROJECT_ID_HOOK_UNIT_TEST

BASE_PACKAGE = "airflow.providers.google.common.hooks.base_google."
SECRETS_HOOK_PACKAGE = "airflow.providers.google.cloud.hooks.secret_manager."
SECRET_ID = "test-secret-id"
REGIONAL_SECRET_LOCATION = "test-location"
SECRET_VERSION = "test-secret-version"


class TestGoogleCloudSecretManagerHook:
    def setup_method(self, method):
        with patch(f"{BASE_PACKAGE}GoogleBaseHook.get_connection", return_value=MagicMock()):
            self.hook = GoogleCloudSecretManagerHook()

    @patch(f"{SECRETS_HOOK_PACKAGE}GoogleCloudSecretManagerHook.client", new_callable=PropertyMock)
    def test__get_parent(self, mock_client):
        expected_value = mock_client.return_value.common_project_path.return_value
        parent = self.hook._get_parent(GCP_PROJECT_ID_HOOK_UNIT_TEST)
        assert expected_value == parent
        mock_client.assert_called_once()
        mock_client.return_value.common_project_path.assert_called_once_with(GCP_PROJECT_ID_HOOK_UNIT_TEST)

        expected_value = mock_client.return_value.common_location_path.return_value
        parent = self.hook._get_parent(GCP_PROJECT_ID_HOOK_UNIT_TEST, REGIONAL_SECRET_LOCATION)
        assert expected_value == parent
        # mock_client.assert_called_once() # will fail as already client has been triggered due to previous case.
        mock_client.return_value.common_location_path.assert_called_once_with(
            GCP_PROJECT_ID_HOOK_UNIT_TEST, REGIONAL_SECRET_LOCATION
        )

    @patch(f"{SECRETS_HOOK_PACKAGE}GoogleCloudSecretManagerHook.client", new_callable=PropertyMock)
    def test__get_secret_path(self, mock_client):
        expected_value = mock_client.return_value.secret_path.return_value
        secret_path = self.hook._get_secret_path(GCP_PROJECT_ID_HOOK_UNIT_TEST, SECRET_ID)
        assert expected_value == secret_path
        mock_client.assert_called_once()
        mock_client.return_value.secret_path.assert_called_once_with(GCP_PROJECT_ID_HOOK_UNIT_TEST, SECRET_ID)

        expected_value = f"projects/{GCP_PROJECT_ID_HOOK_UNIT_TEST}/locations/{REGIONAL_SECRET_LOCATION}/secrets/{SECRET_ID}"
        parent = self.hook._get_secret_path(
            GCP_PROJECT_ID_HOOK_UNIT_TEST, SECRET_ID, REGIONAL_SECRET_LOCATION
        )
        assert expected_value == parent

    @patch(f"{SECRETS_HOOK_PACKAGE}GoogleCloudSecretManagerHook.client", new_callable=PropertyMock)
    def test__get_secret_version_path(self, mock_client):
        expected_value = mock_client.return_value.secret_version_path.return_value
        parent = self.hook._get_secret_version_path(GCP_PROJECT_ID_HOOK_UNIT_TEST, SECRET_ID, SECRET_VERSION)
        assert expected_value == parent
        mock_client.assert_called_once()
        mock_client.return_value.secret_version_path.assert_called_once_with(
            GCP_PROJECT_ID_HOOK_UNIT_TEST, SECRET_ID, SECRET_VERSION
        )

        expected_value = f"projects/{GCP_PROJECT_ID_HOOK_UNIT_TEST}/locations/{REGIONAL_SECRET_LOCATION}/secrets/{SECRET_ID}/versions/{SECRET_VERSION}"
        parent = self.hook._get_secret_version_path(
            GCP_PROJECT_ID_HOOK_UNIT_TEST, SECRET_ID, SECRET_VERSION, REGIONAL_SECRET_LOCATION
        )
        assert expected_value == parent

    @patch(f"{SECRETS_HOOK_PACKAGE}GoogleCloudSecretManagerHook.get_credentials")
    @patch(f"{SECRETS_HOOK_PACKAGE}SecretManagerServiceClient")
    def test_client(self, mock_client, mock_get_credentials):
        mock_client_result = mock_client.return_value
        mock_credentials = self.hook.get_credentials.return_value

        client_1 = self.hook.client
        client_2 = self.hook.client

        assert client_1 == mock_client_result
        assert client_1 == client_2
        mock_client.assert_called_once_with(credentials=mock_credentials, client_info=CLIENT_INFO)
        mock_get_credentials.assert_called_once()

    @patch(f"{SECRETS_HOOK_PACKAGE}GoogleCloudSecretManagerHook.client", new_callable=PropertyMock)
    def test_get_conn(self, mock_client):
        mock_client_result = mock_client.return_value

        client_1 = self.hook.get_conn()

        assert client_1 == mock_client_result
        mock_client.assert_called_once()

    @pytest.mark.parametrize(
        ("input_secret", "expected_secret"),
        [
            (None, {"replication": {"automatic": {}}}),
            (mock_secret := MagicMock(), mock_secret),  # type: ignore[name-defined]
        ],
    )
    @patch(f"{SECRETS_HOOK_PACKAGE}GoogleCloudSecretManagerHook._get_parent")
    @patch(f"{SECRETS_HOOK_PACKAGE}GoogleCloudSecretManagerHook.client", new_callable=PropertyMock)
    def test_create_secret(self, mock_client, mock_get_parent, input_secret, expected_secret):
        expected_parent = mock_get_parent.return_value  # f"projects/{GCP_PROJECT_ID_HOOK_UNIT_TEST}"
        expected_response = mock_client.return_value.create_secret.return_value
        mock_retry, mock_timeout, mock_metadata = MagicMock(), MagicMock(), MagicMock()

        actual_response = self.hook.create_secret(
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
            secret_id=SECRET_ID,
            secret=input_secret,
            retry=mock_retry,
            timeout=mock_timeout,
            metadata=mock_metadata,
        )

        assert actual_response == expected_response
        mock_client.assert_called_once()
        mock_get_parent.assert_called_once_with(GCP_PROJECT_ID_HOOK_UNIT_TEST, None)
        mock_client.return_value.create_secret.assert_called_once_with(
            request={
                "parent": expected_parent,
                "secret_id": SECRET_ID,
                "secret": expected_secret,
            },
            retry=mock_retry,
            timeout=mock_timeout,
            metadata=mock_metadata,
        )

    @patch(f"{SECRETS_HOOK_PACKAGE}GoogleCloudSecretManagerHook._get_secret_path")
    @patch(f"{SECRETS_HOOK_PACKAGE}GoogleCloudSecretManagerHook.client", new_callable=PropertyMock)
    def test_add_secret_version(self, mock_client, mock_get_secret_path):
        expected_parent = mock_get_secret_path.return_value
        expected_response = mock_client.return_value.add_secret_version.return_value
        mock_payload, mock_retry, mock_timeout, mock_metadata = (MagicMock() for _ in range(4))

        actual_response = self.hook.add_secret_version(
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
            secret_id=SECRET_ID,
            secret_payload=mock_payload,
            retry=mock_retry,
            timeout=mock_timeout,
            metadata=mock_metadata,
        )

        assert actual_response == expected_response
        mock_client.assert_called_once()
        mock_get_secret_path.assert_called_once_with(GCP_PROJECT_ID_HOOK_UNIT_TEST, SECRET_ID, None)
        mock_client.return_value.add_secret_version.assert_called_once_with(
            request={
                "parent": expected_parent,
                "payload": mock_payload,
            },
            retry=mock_retry,
            timeout=mock_timeout,
            metadata=mock_metadata,
        )

    @patch(f"{SECRETS_HOOK_PACKAGE}GoogleCloudSecretManagerHook._get_parent")
    @patch(f"{SECRETS_HOOK_PACKAGE}GoogleCloudSecretManagerHook.client", new_callable=PropertyMock)
    def test_list_secrets(self, mock_client, mock_get_parent):
        expected_parent = mock_get_parent.return_value
        expected_response = mock_client.return_value.list_secrets.return_value
        mock_filter, mock_retry, mock_timeout, mock_metadata = (MagicMock() for _ in range(4))
        page_size, page_token = 20, "test-page-token"

        actual_response = self.hook.list_secrets(
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
            secret_filter=mock_filter,
            page_size=page_size,
            page_token=page_token,
            retry=mock_retry,
            timeout=mock_timeout,
            metadata=mock_metadata,
        )

        assert actual_response == expected_response
        mock_client.assert_called_once()
        mock_get_parent.assert_called_once_with(GCP_PROJECT_ID_HOOK_UNIT_TEST, None)
        mock_client.return_value.list_secrets.assert_called_once_with(
            request={
                "parent": expected_parent,
                "page_size": page_size,
                "page_token": page_token,
                "filter": mock_filter,
            },
            retry=mock_retry,
            timeout=mock_timeout,
            metadata=mock_metadata,
        )

    @pytest.mark.parametrize(
        ("secret_names", "secret_id", "secret_exists_expected"),
        [
            ([], SECRET_ID, False),
            (["secret/name"], SECRET_ID, False),
            (["secret/name1", "secret/name1"], SECRET_ID, False),
            ([f"secret/{SECRET_ID}"], SECRET_ID, True),
            ([f"secret/{SECRET_ID}", "secret/name"], SECRET_ID, True),
            (["secret/name", f"secret/{SECRET_ID}"], SECRET_ID, True),
            (["name1", SECRET_ID], SECRET_ID, True),
        ],
    )
    @patch(f"{SECRETS_HOOK_PACKAGE}GoogleCloudSecretManagerHook.client", new_callable=PropertyMock)
    @patch(f"{SECRETS_HOOK_PACKAGE}GoogleCloudSecretManagerHook.list_secrets")
    def test_secret_exists(
        self, mock_list_secrets, mock_client, secret_names, secret_id, secret_exists_expected
    ):
        list_secrets = []
        for secret_name in secret_names:
            mock_secret = MagicMock()
            mock_secret.name = secret_name
            list_secrets.append(mock_secret)
        mock_list_secrets.return_value = list_secrets
        secret_filter = f"name:{secret_id}"

        secret_exists_actual = self.hook.secret_exists(
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST, secret_id=secret_id
        )

        assert secret_exists_actual == secret_exists_expected
        mock_client.return_value.secret_path.assert_called_once_with(GCP_PROJECT_ID_HOOK_UNIT_TEST, secret_id)
        mock_list_secrets.assert_called_once_with(
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
            page_size=100,
            secret_filter=secret_filter,
            location=None,
        )

    @patch(f"{SECRETS_HOOK_PACKAGE}GoogleCloudSecretManagerHook.client", new_callable=PropertyMock)
    def test_access_secret(self, mock_client):
        expected_response = mock_client.return_value.access_secret_version.return_value
        mock_retry, mock_timeout, mock_metadata = (MagicMock() for _ in range(3))
        secret_version = "test-secret-version"
        mock_name = mock_client.return_value.secret_version_path.return_value

        actual_response = self.hook.access_secret(
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
            secret_id=SECRET_ID,
            secret_version=secret_version,
            retry=mock_retry,
            timeout=mock_timeout,
            metadata=mock_metadata,
        )

        assert actual_response == expected_response
        assert mock_client.call_count == 2
        mock_client.return_value.secret_version_path.assert_called_once_with(
            GCP_PROJECT_ID_HOOK_UNIT_TEST, SECRET_ID, secret_version
        )
        mock_client.return_value.access_secret_version.assert_called_once_with(
            request={"name": mock_name},
            retry=mock_retry,
            timeout=mock_timeout,
            metadata=mock_metadata,
        )

    @patch(f"{SECRETS_HOOK_PACKAGE}GoogleCloudSecretManagerHook.client", new_callable=PropertyMock)
    def test_delete_secret(self, mock_client):
        mock_retry, mock_timeout, mock_metadata = (MagicMock() for _ in range(3))
        mock_name = mock_client.return_value.secret_path.return_value

        actual_response = self.hook.delete_secret(
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
            secret_id=SECRET_ID,
            retry=mock_retry,
            timeout=mock_timeout,
            metadata=mock_metadata,
        )

        assert actual_response is None
        assert mock_client.call_count == 2
        mock_client.return_value.secret_path.assert_called_once_with(GCP_PROJECT_ID_HOOK_UNIT_TEST, SECRET_ID)
        mock_client.return_value.delete_secret.assert_called_once_with(
            request={"name": mock_name},
            retry=mock_retry,
            timeout=mock_timeout,
            metadata=mock_metadata,
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains Google Data Fusion links."""

from __future__ import annotations

from airflow.providers.google.cloud.links.base import BaseGoogleLink

BASE_LINK = "https://console.cloud.google.com/data-fusion"
DATAFUSION_INSTANCE_LINK = BASE_LINK + "/locations/{region}/instances/{instance_name}?project={project_id}"
DATAFUSION_PIPELINES_LINK = "{uri}/cdap/ns/{namespace}/pipelines"
DATAFUSION_PIPELINE_LINK = "{uri}/pipelines/ns/{namespace}/view/{pipeline_name}"


class DataFusionInstanceLink(BaseGoogleLink):
    """Helper class for constructing Data Fusion Instance link."""

    name = "Data Fusion Instance"
    key = "instance_conf"
    format_str = DATAFUSION_INSTANCE_LINK


class DataFusionPipelineLink(BaseGoogleLink):
    """Helper class for constructing Data Fusion Pipeline link."""

    name = "Data Fusion Pipeline"
    key = "pipeline_conf"
    format_str = DATAFUSION_PIPELINE_LINK


class DataFusionPipelinesLink(BaseGoogleLink):
    """Helper class for constructing list of Data Fusion Pipelines link."""

    name = "Data Fusion Pipelines List"
    key = "pipelines_conf"
    format_str = DATAFUSION_PIPELINES_LINK



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from unittest import mock

from airflow.providers.google.cloud.hooks.dataprep import JobGroupStatuses
from airflow.providers.google.cloud.sensors.dataprep import DataprepJobGroupIsFinishedSensor

JOB_GROUP_ID = 1312


class TestDataprepJobGroupIsFinishedSensor:
    @mock.patch("airflow.providers.google.cloud.sensors.dataprep.GoogleDataprepHook")
    def test_passing_arguments_to_hook(self, hook_mock):
        sensor = DataprepJobGroupIsFinishedSensor(
            task_id="check_job_group_finished",
            job_group_id=JOB_GROUP_ID,
        )

        hook_mock.return_value.get_job_group_status.return_value = JobGroupStatuses.COMPLETE
        is_job_group_finished = sensor.poke(context=mock.MagicMock())

        assert is_job_group_finished

        hook_mock.assert_called_once_with(
            dataprep_conn_id="dataprep_default",
        )
        hook_mock.return_value.get_job_group_status.assert_called_once_with(
            job_group_id=JOB_GROUP_ID,
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import asyncio
import logging
from asyncio import Future
from unittest import mock

import pytest
from google.cloud.devtools.cloudbuild_v1.types import Build, BuildStep

from airflow.providers.google.cloud.triggers.cloud_build import CloudBuildCreateBuildTrigger
from airflow.triggers.base import TriggerEvent

CLOUD_BUILD_PATH = "airflow.providers.google.cloud.triggers.cloud_build.{}"
TEST_PROJECT_ID = "cloud-build-project"
TEST_BUILD_ID = "test-build-id-9832662"
REPO_SOURCE = {"repo_source": {"repo_name": "test_repo", "branch_name": "main"}}
TEST_BUILD = {
    "source": REPO_SOURCE,
    "steps": [{"name": "gcr.io/cloud-builders/gcloud", "entrypoint": "/bin/sh", "args": ["-c", "ls"]}],
    "status": "SUCCESS",
}
TEST_BUILD_WORKING = {
    "source": REPO_SOURCE,
    "steps": [{"name": "gcr.io/cloud-builders/gcloud", "entrypoint": "/bin/sh", "args": ["-c", "ls"]}],
    "status": "WORKING",
}

TEST_CONN_ID = "google_cloud_default"
TEST_POLL_INTERVAL = 4.0
TEST_LOCATION = "global"
TEST_BUILD_INSTANCE = dict(
    id="test-build-id-9832662",
    status=3,
    steps=[
        {
            "name": "ubuntu",
            "env": [],
            "args": [],
            "dir_": "",
            "id": "",
            "wait_for": [],
            "entrypoint": "",
            "secret_env": [],
            "volumes": [],
            "status": 0,
            "script": "",
            "allow_failure": False,
            "exit_code": 0,
            "allow_exit_codes": [],
        }
    ],
    name="",
    project_id="",
    status_detail="",
    images=[],
    logs_bucket="",
    build_trigger_id="",
    log_url="",
    substitutions={},
    tags=[],
    secrets=[],
    timing={},
    service_account="",
    warnings=[],
    dependencies=[],
)


@pytest.fixture
def trigger():
    return CloudBuildCreateBuildTrigger(
        id_=TEST_BUILD_ID,
        project_id=TEST_PROJECT_ID,
        gcp_conn_id=TEST_CONN_ID,
        impersonation_chain=None,
        poll_interval=TEST_POLL_INTERVAL,
        location=TEST_LOCATION,
    )


class TestCloudBuildCreateBuildTrigger:
    @staticmethod
    def _mock_build_result(result_to_mock):
        f = Future()
        f.set_result(result_to_mock)
        return f

    def test_serialization(self, trigger):
        """
        Asserts that the CloudBuildCreateBuildTrigger correctly serializes its arguments
        and classpath.
        """

        classpath, kwargs = trigger.serialize()
        assert classpath == "airflow.providers.google.cloud.triggers.cloud_build.CloudBuildCreateBuildTrigger"
        assert kwargs == {
            "id_": TEST_BUILD_ID,
            "project_id": TEST_PROJECT_ID,
            "gcp_conn_id": TEST_CONN_ID,
            "impersonation_chain": None,
            "poll_interval": TEST_POLL_INTERVAL,
            "location": TEST_LOCATION,
        }

    @pytest.mark.asyncio
    @mock.patch(CLOUD_BUILD_PATH.format("CloudBuildAsyncHook"))
    async def test_trigger_on_success_yield_successfully(self, mock_hook, trigger):
        """
        Tests the CloudBuildCreateBuildTrigger only fires once the job execution reaches a successful state.
        """
        mock_hook.return_value.get_cloud_build.return_value = self._mock_build_result(
            Build(id=TEST_BUILD_ID, status=Build.Status.SUCCESS, steps=[BuildStep(name="ubuntu")])
        )
        generator = trigger.run()
        actual = await generator.asend(None)
        assert (
            TriggerEvent(
                {
                    "instance": TEST_BUILD_INSTANCE,
                    "id_": TEST_BUILD_ID,
                    "status": "success",
                    "message": "Build completed",
                }
            )
            == actual
        )

    @pytest.mark.asyncio
    @mock.patch(CLOUD_BUILD_PATH.format("CloudBuildAsyncHook"))
    async def test_trigger_on_running_wait_successfully(self, mock_hook, caplog, trigger):
        """
        Test that CloudBuildCreateBuildTrigger does not fire while a build is still running.
        """
        mock_hook.return_value.get_cloud_build.return_value = self._mock_build_result(
            Build(id=TEST_BUILD_ID, status=Build.Status.WORKING, steps=[BuildStep(name="ubuntu")])
        )
        caplog.set_level(logging.INFO)

        task = asyncio.create_task(trigger.run().__anext__())
        await asyncio.sleep(0.5)

        # TriggerEvent was not returned
        assert task.done() is False

        assert "Build is still running..." in caplog.text
        assert f"Sleeping for {TEST_POLL_INTERVAL} seconds." in caplog.text

        # Prevents error when task is destroyed while in "pending" state
        asyncio.get_event_loop().stop()

    @pytest.mark.asyncio
    @mock.patch(CLOUD_BUILD_PATH.format("CloudBuildAsyncHook"))
    async def test_trigger_on_error_yield_successfully(self, mock_hook, caplog, trigger):
        """
        Test that CloudBuildCreateBuildTrigger fires the correct event in case of an error.
        """
        mock_hook.return_value.get_cloud_build.return_value = self._mock_build_result(
            Build(
                id=TEST_BUILD_ID,
                status=Build.Status.FAILURE,
                steps=[BuildStep(name="ubuntu")],
                status_detail="error",
            )
        )
        caplog.set_level(logging.INFO)

        generator = trigger.run()
        actual = await generator.asend(None)
        assert TriggerEvent({"status": "error", "message": "error"}) == actual

    @pytest.mark.asyncio
    @mock.patch(CLOUD_BUILD_PATH.format("CloudBuildAsyncHook"))
    async def test_trigger_on_exec_yield_successfully(self, mock_hook, trigger):
        """
        Test that CloudBuildCreateBuildTrigger fires the correct event in case of an error.
        """
        mock_hook.return_value.get_cloud_build.side_effect = Exception("Test exception")

        generator = trigger.run()
        actual = await generator.asend(None)
        assert TriggerEvent({"status": "error", "message": "Test exception"}) == actual



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from airflow.providers.google.cloud.links.base import BaseGoogleLink

MANAGED_KAFKA_BASE_LINK = "/managedkafka"
MANAGED_KAFKA_CLUSTER_LINK = (
    MANAGED_KAFKA_BASE_LINK + "/{location}/clusters/{cluster_id}?project={project_id}"
)
MANAGED_KAFKA_CLUSTER_LIST_LINK = MANAGED_KAFKA_BASE_LINK + "/clusters?project={project_id}"
MANAGED_KAFKA_TOPIC_LINK = (
    MANAGED_KAFKA_BASE_LINK + "/{location}/clusters/{cluster_id}/topics/{topic_id}?project={project_id}"
)
MANAGED_KAFKA_CONSUMER_GROUP_LINK = (
    MANAGED_KAFKA_BASE_LINK
    + "/{location}/clusters/{cluster_id}/consumer_groups/{consumer_group_id}?project={project_id}"
)


class ApacheKafkaClusterLink(BaseGoogleLink):
    """Helper class for constructing Apache Kafka Cluster link."""

    name = "Apache Kafka Cluster"
    key = "cluster_conf"
    format_str = MANAGED_KAFKA_CLUSTER_LINK


class ApacheKafkaClusterListLink(BaseGoogleLink):
    """Helper class for constructing Apache Kafka Clusters link."""

    name = "Apache Kafka Cluster List"
    key = "cluster_list_conf"
    format_str = MANAGED_KAFKA_CLUSTER_LIST_LINK


class ApacheKafkaTopicLink(BaseGoogleLink):
    """Helper class for constructing Apache Kafka Topic link."""

    name = "Apache Kafka Topic"
    key = "topic_conf"
    format_str = MANAGED_KAFKA_TOPIC_LINK


class ApacheKafkaConsumerGroupLink(BaseGoogleLink):
    """Helper class for constructing Apache Kafka Consumer Group link."""

    name = "Apache Kafka Consumer Group"
    key = "consumer_group_conf"
    format_str = MANAGED_KAFKA_CONSUMER_GROUP_LINK



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

from google.genai.types import (
    Content,
    CreateCachedContentConfig,
    EmbedContentConfig,
    GoogleSearch,
    Part,
    Tool,
)

from airflow.providers.google.cloud.hooks.gen_ai import (
    GenAIGenerativeModelHook,
)

from unit.google.cloud.utils.base_gcp_mock import (
    mock_base_gcp_hook_default_project_id,
)

TEST_GCP_CONN_ID: str = "test-gcp-conn-id"
GCP_PROJECT = "test-project"
GCP_LOCATION = "us-central1"

TEST_PROMPT = "In 10 words or less, what is apache airflow?"
TEST_CONTENTS = [TEST_PROMPT]
TEST_LANGUAGE_PRETRAINED_MODEL = "textembedding-gecko"
TEST_TEMPERATURE = 0.0
TEST_MAX_OUTPUT_TOKENS = 256
TEST_TOP_P = 0.8
TEST_TOP_K = 40

TEST_TEXT_EMBEDDING_MODEL = ""
TEST_TEXT_EMBEDDING_CONFIG = EmbedContentConfig(output_dimensionality=10)

TEST_MULTIMODAL_PRETRAINED_MODEL = "gemini-pro"

TEST_GENERATION_CONFIG = {
    "max_output_tokens": TEST_MAX_OUTPUT_TOKENS,
    "top_p": TEST_TOP_P,
    "temperature": TEST_TEMPERATURE,
}
TEST_TOOLS = [Tool(google_search=GoogleSearch())]

TEST_MULTIMODAL_VISION_MODEL = "gemini-pro-vision"

SOURCE_MODEL = "gemini-1.0-pro-002"
TRAIN_DATASET = "gs://cloud-samples-data/ai-platform/generative_ai/sft_train_data.jsonl"

TEST_CACHED_MODEL = "gemini-1.5-pro-002"
TEST_CACHED_SYSTEM_INSTRUCTION = """
You are an expert researcher. You always stick to the facts in the sources provided, and never make up new facts.
Now look at these research papers, and answer the following questions.
"""

CACHED_CONTENT_CONFIG = CreateCachedContentConfig(
    contents=[
        Content(
            role="user",
            parts=[
                Part.from_uri(
                    file_uri="gs://cloud-samples-data/generative-ai/pdf/2312.11805v3.pdf",
                    mime_type="application/pdf",
                ),
                Part.from_uri(
                    file_uri="gs://cloud-samples-data/generative-ai/pdf/2403.05530.pdf",
                    mime_type="application/pdf",
                ),
            ],
        )
    ],
    system_instruction=TEST_CACHED_SYSTEM_INSTRUCTION,
    display_name="test-cache",
    ttl="3600s",
)

BASE_STRING = "airflow.providers.google.common.hooks.base_google.{}"
GENERATIVE_MODEL_STRING = "airflow.providers.google.cloud.hooks.gen_ai.{}"


def assert_warning(msg: str, warnings):
    assert any(msg in str(w) for w in warnings)


class TestGenAIGenerativeModelHookWithDefaultProjectId:
    def dummy_get_credentials(self):
        pass

    def setup_method(self):
        with mock.patch(
            BASE_STRING.format("GoogleBaseHook.__init__"), new=mock_base_gcp_hook_default_project_id
        ):
            self.hook = GenAIGenerativeModelHook(gcp_conn_id=TEST_GCP_CONN_ID)
            self.hook.get_credentials = self.dummy_get_credentials

    @mock.patch(GENERATIVE_MODEL_STRING.format("GenAIGenerativeModelHook.get_genai_client"))
    def test_text_embedding_model_get_embeddings(self, mock_get_client) -> None:
        client_mock = mock_get_client.return_value
        client_mock.models = mock.Mock()
        self.hook.embed_content(
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            contents=TEST_CONTENTS,
            model=TEST_TEXT_EMBEDDING_MODEL,
            config=TEST_TEXT_EMBEDDING_CONFIG,
        )
        client_mock.models.embed_content.assert_called_once_with(
            model=TEST_TEXT_EMBEDDING_MODEL,
            contents=TEST_CONTENTS,
            config=TEST_TEXT_EMBEDDING_CONFIG,
        )

    @mock.patch(GENERATIVE_MODEL_STRING.format("GenAIGenerativeModelHook.get_genai_client"))
    def test_generative_model_generate_content(self, mock_get_client) -> None:
        client_mock = mock_get_client.return_value
        client_mock.models = mock.Mock()
        self.hook.generate_content(
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            contents=TEST_CONTENTS,
            generation_config=TEST_GENERATION_CONFIG,
            model=TEST_MULTIMODAL_PRETRAINED_MODEL,
        )
        client_mock.models.generate_content.assert_called_once_with(
            model=TEST_MULTIMODAL_PRETRAINED_MODEL,
            contents=TEST_CONTENTS,
            config=TEST_GENERATION_CONFIG,
        )

    @mock.patch(GENERATIVE_MODEL_STRING.format("GenAIGenerativeModelHook.get_genai_client"))
    def test_supervised_fine_tuning_train(self, mock_get_client) -> None:
        client_mock = mock_get_client.return_value
        client_mock.models = mock.Mock()
        self.hook.supervised_fine_tuning_train(
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            source_model=SOURCE_MODEL,
            training_dataset=TRAIN_DATASET,
        )
        client_mock.tunings.tune.assert_called_once_with(
            base_model=SOURCE_MODEL,
            training_dataset=TRAIN_DATASET,
            config=None,
        )

    @mock.patch(GENERATIVE_MODEL_STRING.format("GenAIGenerativeModelHook.get_genai_client"))
    def test_count_tokens(self, mock_get_client) -> None:
        client_mock = mock_get_client.return_value
        client_mock.models = mock.Mock()
        self.hook.count_tokens(
            project_id=GCP_PROJECT,
            contents=TEST_CONTENTS,
            location=GCP_LOCATION,
            model=TEST_MULTIMODAL_PRETRAINED_MODEL,
        )
        client_mock.models.count_tokens.assert_called_once_with(
            model=TEST_MULTIMODAL_PRETRAINED_MODEL,
            contents=TEST_CONTENTS,
            config=None,
        )

    @mock.patch(GENERATIVE_MODEL_STRING.format("GenAIGenerativeModelHook.get_genai_client"))
    def test_create_cached_content(self, mock_get_client) -> None:
        client_mock = mock_get_client.return_value
        client_mock.models = mock.Mock()
        self.hook.create_cached_content(
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            model=TEST_CACHED_MODEL,
            cached_content_config=CACHED_CONTENT_CONFIG,
        )
        client_mock.caches.create.assert_called_once_with(
            model=TEST_CACHED_MODEL,
            config=CACHED_CONTENT_CONFIG,
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock
from unittest.mock import PropertyMock

import pytest
from google.api_core.gapic_v1.method import DEFAULT
from google.cloud.exceptions import NotFound
from google.cloud.memcache_v1beta2.types import cloud_memcache
from google.cloud.redis_v1.types import Instance

from airflow import version
from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.hooks.cloud_memorystore import (
    CloudMemorystoreHook,
    CloudMemorystoreMemcachedHook,
)

from unit.google.cloud.utils.base_gcp_mock import (
    GCP_PROJECT_ID_HOOK_UNIT_TEST,
    mock_base_gcp_hook_default_project_id,
    mock_base_gcp_hook_no_default_project_id,
)

TEST_GCP_CONN_ID = "test-gcp-conn-id"
TEST_LOCATION = "test-location"
TEST_INSTANCE_ID = "test-instance-id"
TEST_PROJECT_ID = "test-project-id"
TEST_RETRY = DEFAULT
TEST_TIMEOUT = 10.0
TEST_METADATA = [("KEY", "VALUE")]
TEST_PAGE_SIZE = 100
TEST_UPDATE_MASK = {"paths": ["memory_size_gb"]}
TEST_UPDATE_MASK_MEMCACHED = {"displayName": "updated name"}
TEST_PARENT = "projects/test-project-id/locations/test-location"
TEST_NAME = "projects/test-project-id/locations/test-location/instances/test-instance-id"
TEST_PARENT_DEFAULT_PROJECT_ID = f"projects/{GCP_PROJECT_ID_HOOK_UNIT_TEST}/locations/test-location"
TEST_NAME_DEFAULT_PROJECT_ID = (
    f"projects/{GCP_PROJECT_ID_HOOK_UNIT_TEST}/locations/test-location/instances/test-instance-id"
)


class TestCloudMemorystoreWithDefaultProjectIdHook:
    def setup_method(self):
        with mock.patch(
            "airflow.providers.google.cloud.hooks.cloud_memorystore.CloudMemorystoreHook.__init__",
            new=mock_base_gcp_hook_default_project_id,
        ):
            self.hook = CloudMemorystoreHook(gcp_conn_id="test")

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_memorystore.CloudMemorystoreHook.get_conn")
    def test_create_instance_when_exists(self, mock_get_conn, mock_project_id):
        mock_get_conn.return_value.get_instance.return_value = Instance(name=TEST_NAME)
        result = self.hook.create_instance(
            location=TEST_LOCATION,
            instance_id=TEST_INSTANCE_ID,
            instance=Instance(name=TEST_NAME),
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_get_conn.return_value.get_instance.assert_called_once_with(
            request=dict(name=TEST_NAME_DEFAULT_PROJECT_ID),
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        assert Instance(name=TEST_NAME) == result

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_memorystore.CloudMemorystoreHook.get_conn")
    def test_create_instance_when_not_exists(self, mock_get_conn, mock_project_id):
        mock_get_conn.return_value.get_instance.side_effect = [
            NotFound("Instance not found"),
            Instance(name=TEST_NAME),
        ]
        mock_get_conn.return_value.create_instance.return_value.result.return_value = Instance(name=TEST_NAME)
        result = self.hook.create_instance(
            location=TEST_LOCATION,
            instance_id=TEST_INSTANCE_ID,
            instance=Instance(name=TEST_NAME),
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_get_conn.return_value.get_instance.assert_has_calls(
            [
                mock.call(
                    request=dict(name=TEST_NAME_DEFAULT_PROJECT_ID),
                    retry=TEST_RETRY,
                    timeout=TEST_TIMEOUT,
                    metadata=TEST_METADATA,
                ),
                mock.call(
                    request=dict(name=TEST_NAME_DEFAULT_PROJECT_ID),
                    retry=TEST_RETRY,
                    timeout=TEST_TIMEOUT,
                    metadata=TEST_METADATA,
                ),
            ]
        )
        mock_get_conn.return_value.create_instance.assert_called_once_with(
            request=dict(
                parent=TEST_PARENT_DEFAULT_PROJECT_ID,
                instance=Instance(
                    name=TEST_NAME,
                    labels={"airflow-version": "v" + version.version.replace(".", "-").replace("+", "-")},
                ),
                instance_id=TEST_INSTANCE_ID,
            ),
            metadata=TEST_METADATA,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
        )
        assert Instance(name=TEST_NAME) == result

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_memorystore.CloudMemorystoreHook.get_conn")
    def test_delete_instance(self, mock_get_conn, mock_project_id):
        self.hook.delete_instance(
            location=TEST_LOCATION,
            instance=TEST_INSTANCE_ID,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_get_conn.return_value.delete_instance.assert_called_once_with(
            request=dict(name=TEST_NAME_DEFAULT_PROJECT_ID),
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_memorystore.CloudMemorystoreHook.get_conn")
    def test_get_instance(self, mock_get_conn, mock_project_id):
        self.hook.get_instance(
            location=TEST_LOCATION,
            instance=TEST_INSTANCE_ID,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_get_conn.return_value.get_instance.assert_called_once_with(
            request=dict(name=TEST_NAME_DEFAULT_PROJECT_ID),
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_memorystore.CloudMemorystoreHook.get_conn")
    def test_list_instances(self, mock_get_conn, mock_project_id):
        self.hook.list_instances(
            location=TEST_LOCATION,
            page_size=TEST_PAGE_SIZE,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_get_conn.return_value.list_instances.assert_called_once_with(
            request=dict(parent=TEST_PARENT_DEFAULT_PROJECT_ID, page_size=TEST_PAGE_SIZE),
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_memorystore.CloudMemorystoreHook.get_conn")
    def test_update_instance(self, mock_get_conn, mock_project_id):
        self.hook.update_instance(
            update_mask=TEST_UPDATE_MASK,
            instance=Instance(name=TEST_NAME),
            location=TEST_LOCATION,
            instance_id=TEST_INSTANCE_ID,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_get_conn.return_value.update_instance.assert_called_once_with(
            request=dict(update_mask=TEST_UPDATE_MASK, instance=Instance(name=TEST_NAME_DEFAULT_PROJECT_ID)),
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )


class TestCloudMemorystoreWithoutDefaultProjectIdHook:
    def setup_method(self):
        with mock.patch(
            "airflow.providers.google.cloud.hooks.cloud_memorystore.CloudMemorystoreHook.__init__",
            new=mock_base_gcp_hook_no_default_project_id,
        ):
            self.hook = CloudMemorystoreHook(gcp_conn_id="test")

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_memorystore.CloudMemorystoreHook.get_conn")
    def test_create_instance_when_exists(self, mock_get_conn):
        mock_get_conn.return_value.get_instance.return_value = Instance(name=TEST_NAME)
        result = self.hook.create_instance(
            location=TEST_LOCATION,
            instance_id=TEST_INSTANCE_ID,
            instance=Instance(name=TEST_NAME),
            project_id=TEST_PROJECT_ID,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_get_conn.return_value.get_instance.assert_called_once_with(
            request=dict(name="projects/test-project-id/locations/test-location/instances/test-instance-id"),
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        assert Instance(name=TEST_NAME) == result

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_memorystore.CloudMemorystoreHook.get_conn")
    def test_create_instance_when_not_exists(self, mock_get_conn):
        mock_get_conn.return_value.get_instance.side_effect = [
            NotFound("Instance not found"),
            Instance(name=TEST_NAME),
        ]
        mock_get_conn.return_value.create_instance.return_value.result.return_value = Instance(name=TEST_NAME)
        result = self.hook.create_instance(
            location=TEST_LOCATION,
            instance_id=TEST_INSTANCE_ID,
            instance=Instance(name=TEST_NAME),
            project_id=TEST_PROJECT_ID,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_get_conn.return_value.get_instance.assert_has_calls(
            [
                mock.call(
                    request=dict(
                        name="projects/test-project-id/locations/test-location/instances/test-instance-id"
                    ),
                    retry=TEST_RETRY,
                    timeout=TEST_TIMEOUT,
                    metadata=TEST_METADATA,
                ),
                mock.call(
                    request=dict(
                        name="projects/test-project-id/locations/test-location/instances/test-instance-id"
                    ),
                    retry=TEST_RETRY,
                    timeout=TEST_TIMEOUT,
                    metadata=TEST_METADATA,
                ),
            ]
        )

        mock_get_conn.return_value.create_instance.assert_called_once_with(
            request=dict(
                parent=TEST_PARENT,
                instance=Instance(
                    name=TEST_NAME,
                    labels={"airflow-version": "v" + version.version.replace(".", "-").replace("+", "-")},
                ),
                instance_id=TEST_INSTANCE_ID,
            ),
            metadata=TEST_METADATA,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
        )
        assert Instance(name=TEST_NAME) == result

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=None,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_memorystore.CloudMemorystoreHook.get_conn")
    def test_create_instance_without_project_id(self, mock_get_conn, mock_project_id):
        with pytest.raises(AirflowException):
            self.hook.create_instance(
                location=TEST_LOCATION,
                instance_id=TEST_INSTANCE_ID,
                instance=Instance(name=TEST_NAME),
                project_id=None,
                retry=TEST_RETRY,
                timeout=TEST_TIMEOUT,
                metadata=TEST_METADATA,
            )

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_memorystore.CloudMemorystoreHook.get_conn")
    def test_delete_instance(self, mock_get_conn):
        self.hook.delete_instance(
            location=TEST_LOCATION,
            instance=TEST_INSTANCE_ID,
            project_id=TEST_PROJECT_ID,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_get_conn.return_value.delete_instance.assert_called_once_with(
            request=dict(name=TEST_NAME), retry=TEST_RETRY, timeout=TEST_TIMEOUT, metadata=TEST_METADATA
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=None,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_memorystore.CloudMemorystoreHook.get_conn")
    def test_delete_instance_without_project_id(self, mock_get_conn, mock_project_id):
        with pytest.raises(AirflowException):
            self.hook.delete_instance(
                location=TEST_LOCATION,
                instance=Instance(name=TEST_NAME),
                project_id=None,
                retry=TEST_RETRY,
                timeout=TEST_TIMEOUT,
                metadata=TEST_METADATA,
            )

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_memorystore.CloudMemorystoreHook.get_conn")
    def test_get_instance(self, mock_get_conn):
        self.hook.get_instance(
            location=TEST_LOCATION,
            instance=TEST_INSTANCE_ID,
            project_id=TEST_PROJECT_ID,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_get_conn.return_value.get_instance.assert_called_once_with(
            request=dict(name=TEST_NAME), retry=TEST_RETRY, timeout=TEST_TIMEOUT, metadata=TEST_METADATA
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=None,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_memorystore.CloudMemorystoreHook.get_conn")
    def test_get_instance_without_project_id(self, mock_get_conn, mock_project_id):
        with pytest.raises(AirflowException):
            self.hook.get_instance(
                location=TEST_LOCATION,
                instance=Instance(name=TEST_NAME),
                project_id=None,
                retry=TEST_RETRY,
                timeout=TEST_TIMEOUT,
                metadata=TEST_METADATA,
            )

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_memorystore.CloudMemorystoreHook.get_conn")
    def test_list_instances(self, mock_get_conn):
        self.hook.list_instances(
            location=TEST_LOCATION,
            page_size=TEST_PAGE_SIZE,
            project_id=TEST_PROJECT_ID,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_get_conn.return_value.list_instances.assert_called_once_with(
            request=dict(parent=TEST_PARENT, page_size=TEST_PAGE_SIZE),
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=None,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_memorystore.CloudMemorystoreHook.get_conn")
    def test_list_instances_without_project_id(self, mock_get_conn, mock_project_id):
        with pytest.raises(AirflowException):
            self.hook.list_instances(
                location=TEST_LOCATION,
                page_size=TEST_PAGE_SIZE,
                project_id=None,
                retry=TEST_RETRY,
                timeout=TEST_TIMEOUT,
                metadata=TEST_METADATA,
            )

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_memorystore.CloudMemorystoreHook.get_conn")
    def test_update_instance(self, mock_get_conn):
        self.hook.update_instance(
            update_mask=TEST_UPDATE_MASK,
            instance=Instance(name=TEST_NAME),
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
            project_id=TEST_PROJECT_ID,
        )
        mock_get_conn.return_value.update_instance.assert_called_once_with(
            request=dict(update_mask={"paths": ["memory_size_gb"]}, instance=Instance(name=TEST_NAME)),
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=None,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_memorystore.CloudMemorystoreHook.get_conn")
    def test_update_instance_without_project_id(self, mock_get_conn, mock_project_id):
        with pytest.raises(AirflowException):
            self.hook.update_instance(
                update_mask=TEST_UPDATE_MASK,
                instance=Instance(name=TEST_NAME),
                retry=TEST_RETRY,
                timeout=TEST_TIMEOUT,
                metadata=TEST_METADATA,
            )


class TestCloudMemorystoreMemcachedWithDefaultProjectIdHook:
    def setup_method(self):
        with mock.patch(
            "airflow.providers.google.cloud.hooks.cloud_memorystore.CloudMemorystoreMemcachedHook.__init__",
            new=mock_base_gcp_hook_default_project_id,
        ):
            self.hook = CloudMemorystoreMemcachedHook(gcp_conn_id="test")

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch(
        "airflow.providers.google.cloud.hooks.cloud_memorystore.CloudMemorystoreMemcachedHook.get_conn"
    )
    def test_create_instance_when_exists(self, mock_get_conn, mock_project_id):
        mock_get_conn.return_value.get_instance.return_value = cloud_memcache.Instance(name=TEST_NAME)
        result = self.hook.create_instance(
            location=TEST_LOCATION,
            instance_id=TEST_INSTANCE_ID,
            instance=cloud_memcache.Instance(name=TEST_NAME),
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_get_conn.return_value.get_instance.assert_called_once_with(
            name=TEST_NAME_DEFAULT_PROJECT_ID, retry=TEST_RETRY, timeout=TEST_TIMEOUT, metadata=TEST_METADATA
        )
        assert cloud_memcache.Instance(name=TEST_NAME) == result

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch(
        "airflow.providers.google.cloud.hooks.cloud_memorystore.CloudMemorystoreMemcachedHook.get_conn"
    )
    def test_create_instance_when_not_exists(self, mock_get_conn, mock_project_id):
        mock_get_conn.return_value.get_instance.side_effect = [
            NotFound("Instance not found"),
            cloud_memcache.Instance(name=TEST_NAME),
        ]
        mock_get_conn.return_value.create_instance.return_value.result.return_value = cloud_memcache.Instance(
            name=TEST_NAME
        )
        result = self.hook.create_instance(
            location=TEST_LOCATION,
            instance_id=TEST_INSTANCE_ID,
            instance=cloud_memcache.Instance(name=TEST_NAME),
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_get_conn.return_value.get_instance.assert_has_calls(
            [
                mock.call(
                    name=TEST_NAME_DEFAULT_PROJECT_ID,
                    retry=TEST_RETRY,
                    timeout=TEST_TIMEOUT,
                    metadata=TEST_METADATA,
                ),
                mock.call(
                    name=TEST_NAME_DEFAULT_PROJECT_ID,
                    retry=TEST_RETRY,
                    timeout=TEST_TIMEOUT,
                    metadata=TEST_METADATA,
                ),
            ]
        )
        mock_get_conn.return_value.create_instance.assert_called_once_with(
            resource=cloud_memcache.Instance(
                name=TEST_NAME,
                labels={"airflow-version": "v" + version.version.replace(".", "-").replace("+", "-")},
            ),
            instance_id=TEST_INSTANCE_ID,
            metadata=TEST_METADATA,
            parent=TEST_PARENT_DEFAULT_PROJECT_ID,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
        )
        assert cloud_memcache.Instance(name=TEST_NAME) == result

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch(
        "airflow.providers.google.cloud.hooks.cloud_memorystore.CloudMemorystoreMemcachedHook.get_conn"
    )
    def test_delete_instance(self, mock_get_conn, mock_project_id):
        self.hook.delete_instance(
            location=TEST_LOCATION,
            instance=TEST_INSTANCE_ID,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_get_conn.return_value.delete_instance.assert_called_once_with(
            name=TEST_NAME_DEFAULT_PROJECT_ID, retry=TEST_RETRY, timeout=TEST_TIMEOUT, metadata=TEST_METADATA
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch(
        "airflow.providers.google.cloud.hooks.cloud_memorystore.CloudMemorystoreMemcachedHook.get_conn"
    )
    def test_get_instance(self, mock_get_conn, mock_project_id):
        self.hook.get_instance(
            location=TEST_LOCATION,
            instance=TEST_INSTANCE_ID,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_get_conn.return_value.get_instance.assert_called_once_with(
            name=TEST_NAME_DEFAULT_PROJECT_ID, retry=TEST_RETRY, timeout=TEST_TIMEOUT, metadata=TEST_METADATA
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch(
        "airflow.providers.google.cloud.hooks.cloud_memorystore.CloudMemorystoreMemcachedHook.get_conn"
    )
    def test_list_instances(self, mock_get_conn, mock_project_id):
        self.hook.list_instances(
            location=TEST_LOCATION,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_get_conn.return_value.list_instances.assert_called_once_with(
            parent=TEST_PARENT_DEFAULT_PROJECT_ID,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch(
        "airflow.providers.google.cloud.hooks.cloud_memorystore.CloudMemorystoreMemcachedHook.get_conn"
    )
    def test_update_instance(self, mock_get_conn, mock_project_id):
        self.hook.update_instance(
            update_mask=TEST_UPDATE_MASK_MEMCACHED,
            instance=cloud_memcache.Instance(name=TEST_NAME),
            location=TEST_LOCATION,
            instance_id=TEST_INSTANCE_ID,
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )
        mock_get_conn.return_value.update_instance.assert_called_once_with(
            update_mask=TEST_UPDATE_MASK_MEMCACHED,
            resource=cloud_memcache.Instance(name=TEST_NAME_DEFAULT_PROJECT_ID),
            retry=TEST_RETRY,
            timeout=TEST_TIMEOUT,
            metadata=TEST_METADATA,
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
This module contains various unit tests for GCP Cloud Build Operators
"""

from __future__ import annotations

import json
import tempfile
from copy import deepcopy
from unittest import mock

import pytest
from google.api_core.gapic_v1.method import DEFAULT
from google.cloud.devtools.cloudbuild_v1.types import Build, BuildTrigger, RepoSource, StorageSource

from airflow.exceptions import AirflowException, TaskDeferred
from airflow.models import DAG
from airflow.models.dagrun import DagRun
from airflow.models.taskinstance import TaskInstance
from airflow.providers.google.cloud.operators.cloud_build import (
    BuildProcessor,
    CloudBuildCancelBuildOperator,
    CloudBuildCreateBuildOperator,
    CloudBuildCreateBuildTriggerOperator,
    CloudBuildDeleteBuildTriggerOperator,
    CloudBuildGetBuildOperator,
    CloudBuildGetBuildTriggerOperator,
    CloudBuildListBuildsOperator,
    CloudBuildListBuildTriggersOperator,
    CloudBuildRetryBuildOperator,
    CloudBuildRunBuildTriggerOperator,
    CloudBuildUpdateBuildTriggerOperator,
)
from airflow.providers.google.cloud.triggers.cloud_build import CloudBuildCreateBuildTrigger
from airflow.utils.timezone import datetime
from airflow.utils.types import DagRunType

GCP_CONN_ID = "google_cloud_default"
PROJECT_ID = "cloud-build-project"
CLOUD_BUILD_HOOK_PATH = "airflow.providers.google.cloud.operators.cloud_build.CloudBuildHook"
BUILD_ID = "test-build-id-9832661"
REPO_SOURCE = {"repo_source": {"repo_name": "test_repo", "branch_name": "main"}}
BUILD = {
    "source": REPO_SOURCE,
    "steps": [{"name": "gcr.io/cloud-builders/gcloud", "entrypoint": "/bin/sh", "args": ["-c", "ls"]}],
    "status": "SUCCESS",
}
BUILD_TRIGGER = {
    "name": "test-cloud-build-trigger",
    "trigger_template": {"project_id": PROJECT_ID, "repo_name": "test_repo", "branch_name": "master"},
    "filename": "cloudbuild.yaml",
}
OPERATION = {"metadata": {"build": {"id": BUILD_ID}}}
TRIGGER_ID = "32488e7f-09d6-4fe9-a5fb-4ca1419a6e7a"
TEST_BUILD_INSTANCE = dict(
    id="test-build-id-9832662",
    status=3,
    steps=[
        {
            "name": "ubuntu",
            "env": [],
            "args": [],
            "dir_": "",
            "id": "",
            "wait_for": [],
            "entrypoint": "",
            "secret_env": [],
            "volumes": [],
            "status": 0,
            "script": "",
        }
    ],
    name="",
    project_id="",
    status_detail="",
    images=[],
    logs_bucket="",
    build_trigger_id="",
    log_url="",
    substitutions={},
    tags=[],
    secrets=[],
    timing={},
    service_account="",
    warnings=[],
)


class TestCloudBuildOperator:
    @mock.patch(CLOUD_BUILD_HOOK_PATH)
    def test_cancel_build(self, mock_hook):
        mock_hook.return_value.cancel_build.return_value = Build()

        operator = CloudBuildCancelBuildOperator(id_=TRIGGER_ID, task_id="id")
        operator.execute(context=mock.MagicMock())

        mock_hook.assert_called_once_with(gcp_conn_id=GCP_CONN_ID, impersonation_chain=None)
        mock_hook.return_value.cancel_build.assert_called_once_with(
            id_=TRIGGER_ID, project_id=None, retry=DEFAULT, timeout=None, metadata=(), location="global"
        )

    @mock.patch(CLOUD_BUILD_HOOK_PATH)
    def test_create_build(self, mock_hook):
        mock_hook.return_value.create_build_without_waiting_for_result.return_value = (BUILD, BUILD_ID)
        mock_hook.return_value.wait_for_operation.return_value = Build()

        operator = CloudBuildCreateBuildOperator(build=BUILD, task_id="id")
        operator.execute(context=mock.MagicMock())

        mock_hook.assert_called_once_with(gcp_conn_id=GCP_CONN_ID, impersonation_chain=None)
        build = Build(BUILD)
        mock_hook.return_value.create_build_without_waiting_for_result.assert_called_once_with(
            build=build, project_id=None, retry=DEFAULT, timeout=None, metadata=(), location="global"
        )
        mock_hook.return_value.wait_for_operation.assert_called_once_with(timeout=None, operation=BUILD)

    @mock.patch(CLOUD_BUILD_HOOK_PATH)
    def test_create_build_with_missing_build(self, mock_hook):
        mock_hook.return_value.create_build_without_waiting_for_result.return_value = Build()
        with pytest.raises((TypeError, AirflowException), match="missing keyword argument 'build'"):
            CloudBuildCreateBuildOperator(task_id="id")

    @pytest.mark.parametrize(
        ("file_type", "file_content"),
        [
            (
                ".json",
                json.dumps({"steps": [{"name": "ubuntu", "args": ["echo", "Hello {{ params.name }}!"]}]}),
            ),
            (
                ".yaml",
                """
                steps:
                - name: 'ubuntu'
                  args: ['echo', 'Hello {{ params.name }}!']
                """,
            ),
        ],
    )
    def test_load_templated(self, file_type, file_content):
        with tempfile.NamedTemporaryFile(suffix=file_type, mode="w+") as f:
            f.writelines(file_content)
            f.flush()

            operator = CloudBuildCreateBuildOperator(
                build=f.name, task_id="task-id", params={"name": "airflow"}
            )
            operator.prepare_template()
            expected_body = {"steps": [{"name": "ubuntu", "args": ["echo", "Hello {{ params.name }}!"]}]}
            assert expected_body == operator.build

    @mock.patch(CLOUD_BUILD_HOOK_PATH)
    def test_create_build_trigger(self, mock_hook):
        mock_hook.return_value.create_build_trigger.return_value = BuildTrigger()

        operator = CloudBuildCreateBuildTriggerOperator(trigger=BUILD_TRIGGER, task_id="id")
        operator.execute(context=mock.MagicMock())

        mock_hook.assert_called_once_with(gcp_conn_id=GCP_CONN_ID, impersonation_chain=None)
        mock_hook.return_value.create_build_trigger.assert_called_once_with(
            trigger=BUILD_TRIGGER,
            project_id=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
            location="global",
        )

    @mock.patch(CLOUD_BUILD_HOOK_PATH)
    def test_delete_build_trigger(self, mock_hook):
        mock_hook.return_value.delete_build_trigger.return_value = None

        operator = CloudBuildDeleteBuildTriggerOperator(trigger_id=TRIGGER_ID, task_id="id")
        operator.execute(context=mock.MagicMock())

        mock_hook.assert_called_once_with(gcp_conn_id=GCP_CONN_ID, impersonation_chain=None)
        mock_hook.return_value.delete_build_trigger.assert_called_once_with(
            trigger_id=TRIGGER_ID,
            project_id=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
            location="global",
        )

    @mock.patch(CLOUD_BUILD_HOOK_PATH)
    def test_get_build(self, mock_hook):
        mock_hook.return_value.get_build.return_value = Build()

        operator = CloudBuildGetBuildOperator(id_=BUILD_ID, task_id="id")
        operator.execute(context=mock.MagicMock())

        mock_hook.assert_called_once_with(gcp_conn_id=GCP_CONN_ID, impersonation_chain=None)
        mock_hook.return_value.get_build.assert_called_once_with(
            id_=BUILD_ID, project_id=None, retry=DEFAULT, timeout=None, metadata=(), location="global"
        )

    @mock.patch(CLOUD_BUILD_HOOK_PATH)
    def test_get_build_trigger(self, mock_hook):
        mock_hook.return_value.get_build_trigger.return_value = BuildTrigger()

        operator = CloudBuildGetBuildTriggerOperator(trigger_id=TRIGGER_ID, task_id="id")
        operator.execute(context=mock.MagicMock())

        mock_hook.assert_called_once_with(gcp_conn_id=GCP_CONN_ID, impersonation_chain=None)
        mock_hook.return_value.get_build_trigger.assert_called_once_with(
            trigger_id=TRIGGER_ID,
            project_id=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
            location="global",
        )

    @mock.patch(CLOUD_BUILD_HOOK_PATH)
    def test_list_build_triggers(self, mock_hook):
        mock_hook.return_value.list_build_triggers.return_value = mock.MagicMock()

        operator = CloudBuildListBuildTriggersOperator(task_id="id", location="global")
        operator.execute(context=mock.MagicMock())

        mock_hook.assert_called_once_with(gcp_conn_id=GCP_CONN_ID, impersonation_chain=None)
        mock_hook.return_value.list_build_triggers.assert_called_once_with(
            project_id=None,
            location="global",
            page_size=None,
            page_token=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(CLOUD_BUILD_HOOK_PATH)
    def test_list_builds(self, mock_hook):
        mock_hook.return_value.list_builds.return_value = mock.MagicMock()

        operator = CloudBuildListBuildsOperator(task_id="id", location="global")
        operator.execute(context=mock.MagicMock())

        mock_hook.assert_called_once_with(gcp_conn_id=GCP_CONN_ID, impersonation_chain=None)
        mock_hook.return_value.list_builds.assert_called_once_with(
            project_id=None,
            location="global",
            page_size=None,
            filter_=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(CLOUD_BUILD_HOOK_PATH)
    def test_retry_build(self, mock_hook):
        mock_hook.return_value.retry_build.return_value = Build()

        operator = CloudBuildRetryBuildOperator(id_=BUILD_ID, task_id="id")
        operator.execute(context=mock.MagicMock())

        mock_hook.assert_called_once_with(gcp_conn_id=GCP_CONN_ID, impersonation_chain=None)
        mock_hook.return_value.retry_build.assert_called_once_with(
            id_=BUILD_ID,
            project_id=None,
            wait=True,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
            location="global",
        )

    @mock.patch(CLOUD_BUILD_HOOK_PATH)
    def test_run_build_trigger(self, mock_hook):
        mock_hook.return_value.run_build_trigger.return_value = Build()

        operator = CloudBuildRunBuildTriggerOperator(trigger_id=TRIGGER_ID, source=REPO_SOURCE, task_id="id")
        operator.execute(context=mock.MagicMock())

        mock_hook.assert_called_once_with(gcp_conn_id=GCP_CONN_ID, impersonation_chain=None)
        mock_hook.return_value.run_build_trigger.assert_called_once_with(
            trigger_id=TRIGGER_ID,
            source=REPO_SOURCE,
            project_id=None,
            wait=True,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
            location="global",
        )

    @mock.patch(CLOUD_BUILD_HOOK_PATH)
    def test_update_build_trigger(self, mock_hook):
        mock_hook.return_value.update_build_trigger.return_value = BuildTrigger()

        operator = CloudBuildUpdateBuildTriggerOperator(
            trigger_id=TRIGGER_ID, trigger=BUILD_TRIGGER, task_id="id"
        )
        operator.execute(context=mock.MagicMock())

        mock_hook.assert_called_once_with(gcp_conn_id=GCP_CONN_ID, impersonation_chain=None)
        mock_hook.return_value.update_build_trigger.assert_called_once_with(
            trigger_id=TRIGGER_ID,
            trigger=BUILD_TRIGGER,
            project_id=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
            location="global",
        )


class TestBuildProcessor:
    def test_verify_source(self):
        error_message = r"The source could not be determined."
        with pytest.raises(AirflowException, match=error_message):
            BuildProcessor(build={"source": {"storage_source": {}, "repo_source": {}}}).process_body()

        with pytest.raises(AirflowException, match=error_message):
            BuildProcessor(build={"source": {}}).process_body()

    @pytest.mark.parametrize(
        ("url", "expected_dict"),
        [
            (
                "https://source.cloud.google.com/airflow-project/airflow-repo",
                {"project_id": "airflow-project", "repo_name": "airflow-repo", "branch_name": "master"},
            ),
            (
                "https://source.cloud.google.com/airflow-project/airflow-repo/+/branch-name:",
                {"project_id": "airflow-project", "repo_name": "airflow-repo", "branch_name": "branch-name"},
            ),
            (
                "https://source.cloud.google.com/airflow-project/airflow-repo/+/feature/branch:",
                {
                    "project_id": "airflow-project",
                    "repo_name": "airflow-repo",
                    "branch_name": "feature/branch",
                },
            ),
        ],
    )
    def test_convert_repo_url_to_dict_valid(self, url, expected_dict):
        body = {"source": {"repo_source": url}}
        body = BuildProcessor(build=body).process_body()
        assert body.source.repo_source == RepoSource(expected_dict)

    @pytest.mark.parametrize(
        "url",
        [
            "http://source.e.com/airflow-project/airflow-repo/branch-name",
            "httpXs://source.cloud.google.com/airflow-project/airflow-repo",
            "://source.cloud.google.com/airflow-project/airflow-repo",
        ],
    )
    def test_convert_repo_url_to_dict_invalid(self, url):
        body = {"source": {"repo_source": url}}
        with pytest.raises(AirflowException, match="Invalid URL."):
            BuildProcessor(build=body).process_body()

    @pytest.mark.parametrize(
        ("url", "expected_dict"),
        [
            (
                "gs://bucket-name/airflow-object.tar.gz",
                {"bucket": "bucket-name", "object_": "airflow-object.tar.gz"},
            ),
            (
                "gs://bucket-name/airflow-object.tar.gz#1231231",
                {"bucket": "bucket-name", "object_": "airflow-object.tar.gz", "generation": 1231231},
            ),
        ],
    )
    def test_convert_storage_url_to_dict_valid(self, url, expected_dict):
        body = {"source": {"storage_source": url}}
        body = BuildProcessor(build=body).process_body()
        assert body.source.storage_source == StorageSource(expected_dict)

    @pytest.mark.parametrize("url", ["///object", "gsXXa:///object", "gs://bucket-name/", "gs://bucket-name"])
    def test_convert_storage_url_to_dict_invalid(self, url):
        body = {"source": {"storage_source": url}}
        with pytest.raises(AirflowException, match="Invalid URL."):
            BuildProcessor(build=body).process_body()

    @pytest.mark.parametrize("source_key", ["storage_source", "repo_source"])
    def test_do_nothing(self, source_key):
        body = {"source": {source_key: {}}}
        expected_body = deepcopy(body)

        BuildProcessor(build=body).process_body()
        assert body == expected_body


def set_execute_complete(session, ti, **next_kwargs):
    ti.next_method = "execute_complete"
    ti.next_kwargs = next_kwargs
    session.flush()


@pytest.mark.db_test
@mock.patch(CLOUD_BUILD_HOOK_PATH)
def test_async_create_build_fires_correct_trigger_should_execute_successfully(
    mock_hook, create_task_instance_of_operator, session
):
    mock_hook.return_value.create_build_without_waiting_for_result.return_value = (BUILD, BUILD_ID)

    ti = create_task_instance_of_operator(
        build=BUILD,
        task_id="id",
        deferrable=True,
        operator_class=CloudBuildCreateBuildOperator,
        dag_id="cloud_build_create_build_job",
    )

    with pytest.raises(TaskDeferred) as exc:
        ti.task.execute({"ti": ti, "task_instance": ti})

    assert isinstance(exc.value.trigger, CloudBuildCreateBuildTrigger), (
        "Trigger is not a CloudBuildCreateBuildTrigger"
    )


@mock.patch(CLOUD_BUILD_HOOK_PATH)
def test_async_create_build_without_wait_should_execute_successfully(mock_hook):
    mock_hook.return_value.create_build_without_waiting_for_result.return_value = (BUILD, BUILD_ID)
    mock_hook.return_value.get_build.return_value = Build()

    operator = CloudBuildCreateBuildOperator(
        build=BUILD,
        task_id="id",
        wait=False,
        deferrable=True,
    )
    operator.execute(context=mock.MagicMock())

    mock_hook.assert_called_once_with(gcp_conn_id=GCP_CONN_ID, impersonation_chain=None)
    build = Build(BUILD)
    mock_hook.return_value.create_build_without_waiting_for_result.assert_called_once_with(
        build=build, project_id=None, retry=DEFAULT, timeout=None, metadata=(), location="global"
    )
    mock_hook.return_value.get_build.assert_called_once_with(id_=BUILD_ID, project_id=None, location="global")


@pytest.mark.db_test
@mock.patch(CLOUD_BUILD_HOOK_PATH)
def test_async_create_build_correct_logging_should_execute_successfully(
    mock_hook, create_task_instance_of_operator, session
):
    mock_hook.return_value.create_build_without_waiting_for_result.return_value = (BUILD, BUILD_ID)
    mock_hook.return_value.get_build.return_value = Build()

    ti = create_task_instance_of_operator(
        build=BUILD,
        task_id="id",
        deferrable=True,
        operator_class=CloudBuildCreateBuildOperator,
        dag_id="cloud_build_create_build_logging",
    )

    with mock.patch.object(ti.task.log, "info") as mock_log_info:
        ti.task.execute_complete(
            context={"ti": ti, "task": ti.task},
            event={
                "instance": TEST_BUILD_INSTANCE,
                "status": "success",
                "message": "Build completed",
                "id_": BUILD_ID,
            },
        )

    mock_log_info.assert_called_with("Cloud Build completed with response %s ", "Build completed")


def test_async_create_build_error_event_should_throw_exception():
    operator = CloudBuildCreateBuildOperator(
        build=BUILD,
        task_id="id",
        deferrable=True,
    )
    with pytest.raises(AirflowException):
        operator.execute_complete(context=None, event={"status": "error", "message": "test failure message"})


@mock.patch(CLOUD_BUILD_HOOK_PATH)
def test_async_create_build_with_missing_build_should_throw_exception(mock_hook):
    mock_hook.return_value.create_build.return_value = Build()
    with pytest.raises((TypeError, AirflowException), match="missing keyword argument 'build'"):
        CloudBuildCreateBuildOperator(task_id="id")


@pytest.mark.parametrize(
    ("file_type", "file_content"),
    [
        (
            ".json",
            json.dumps({"steps": [{"name": "ubuntu", "args": ["echo", "Hello {{ params.name }}!"]}]}),
        ),
        (
            ".yaml",
            """
            steps:
            - name: 'ubuntu'
              args: ['echo', 'Hello {{ params.name }}!']
            """,
        ),
    ],
)
def test_async_load_templated_should_execute_successfully(file_type, file_content):
    with tempfile.NamedTemporaryFile(suffix=file_type, mode="w+") as f:
        f.writelines(file_content)
        f.flush()

        operator = CloudBuildCreateBuildOperator(
            build=f.name,
            task_id="task-id",
            params={"name": "airflow"},
            deferrable=True,
        )
        operator.prepare_template()
        expected_body = {"steps": [{"name": "ubuntu", "args": ["echo", "Hello {{ params.name }}!"]}]}
        assert expected_body == operator.build


def create_context(task):
    dag = DAG(dag_id="dag", schedule=None)
    logical_date = datetime(2022, 1, 1, 0, 0, 0)
    dag_run = DagRun(
        dag_id=dag.dag_id,
        logical_date=logical_date,
        run_id=DagRun.generate_run_id(DagRunType.MANUAL, logical_date),
    )
    task_instance = TaskInstance(task=task)
    task_instance.dag_run = dag_run
    task_instance.dag_id = dag.dag_id
    task_instance.xcom_push = mock.Mock()
    return {
        "dag": dag,
        "run_id": dag_run.run_id,
        "task": task,
        "ti": task_instance,
        "task_instance": task_instance,
        "logical_date": logical_date,
    }



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock
from unittest.mock import MagicMock

import pytest
from psycopg2.extras import Json

from airflow.providers.google.cloud.transfers.bigquery_to_postgres import BigQueryToPostgresOperator

TASK_ID = "test-bq-create-table-operator"
TEST_DATASET = "test-dataset"
TEST_TABLE_ID = "test-table-id"
TEST_DAG_ID = "test-bigquery-operators"
TEST_DESTINATION_TABLE = "table"
TEST_PROJECT = "test-project"


def _make_bq_table(schema_names: list[str]):
    class TableObj:
        def __init__(self, schema):
            self.schema = []
            for n in schema:
                field = MagicMock()
                field.name = n
                self.schema.append(field)
            self.description = "table description"
            self.external_data_configuration = None
            self.labels = {}
            self.num_rows = 0
            self.num_bytes = 0
            self.table_type = "TABLE"

    return TableObj(schema_names)


class TestBigQueryToPostgresOperator:
    @mock.patch("airflow.providers.google.cloud.transfers.bigquery_to_postgres.bigquery_get_data")
    @mock.patch.object(BigQueryToPostgresOperator, "bigquery_hook", new_callable=mock.PropertyMock)
    @mock.patch.object(BigQueryToPostgresOperator, "postgres_hook", new_callable=mock.PropertyMock)
    def test_execute_good_request_to_bq(self, mock_pg_hook, mock_bq_hook, mock_bigquery_get_data):
        operator = BigQueryToPostgresOperator(
            task_id=TASK_ID,
            dataset_table=f"{TEST_DATASET}.{TEST_TABLE_ID}",
            target_table_name=TEST_DESTINATION_TABLE,
            replace=False,
        )

        mock_bigquery_get_data.return_value = [[("row1", "val1")], [("row2", "val2")]]
        mock_pg = mock.MagicMock()
        mock_pg_hook.return_value = mock_pg
        mock_bq = mock.MagicMock()
        mock_bq.project_id = TEST_PROJECT
        mock_bq_hook.return_value = mock_bq

        operator.execute(context=mock.MagicMock())

        mock_bigquery_get_data.assert_called_once_with(
            operator.log,
            TEST_DATASET,
            TEST_TABLE_ID,
            mock_bq,
            operator.batch_size,
            operator.selected_fields,
        )
        assert mock_pg.insert_rows.call_count == 2

    @mock.patch("airflow.providers.google.cloud.transfers.bigquery_to_postgres.bigquery_get_data")
    @mock.patch.object(BigQueryToPostgresOperator, "bigquery_hook", new_callable=mock.PropertyMock)
    @mock.patch.object(BigQueryToPostgresOperator, "postgres_hook", new_callable=mock.PropertyMock)
    def test_execute_good_request_to_bq_with_replace(
        self,
        mock_pg_hook,
        mock_bq_hook,
        mock_bigquery_get_data,
    ):
        operator = BigQueryToPostgresOperator(
            task_id=TASK_ID,
            dataset_table=f"{TEST_DATASET}.{TEST_TABLE_ID}",
            target_table_name=TEST_DESTINATION_TABLE,
            replace=True,
            selected_fields=["col_1", "col_2"],
            replace_index=["col_1"],
        )

        mock_bigquery_get_data.return_value = [[("only_row", "val")]]
        mock_pg = mock.MagicMock()
        mock_pg_hook.return_value = mock_pg
        mock_bq = mock.MagicMock()
        mock_bq.project_id = TEST_PROJECT
        mock_bq_hook.return_value = mock_bq

        operator.execute(context=mock.MagicMock())

        mock_bigquery_get_data.assert_called_once_with(
            operator.log,
            TEST_DATASET,
            TEST_TABLE_ID,
            mock_bq,
            operator.batch_size,
            ["col_1", "col_2"],
        )
        mock_pg.insert_rows.assert_called_once_with(
            table=TEST_DESTINATION_TABLE,
            rows=[("only_row", "val")],
            target_fields=["col_1", "col_2"],
            replace=True,
            commit_every=operator.batch_size,
            replace_index=["col_1"],
        )

    @pytest.mark.parametrize(
        ("selected_fields", "replace_index"),
        [(None, None), (["col_1, col_2"], None), (None, ["col_1"])],
    )
    def test_init_raises_exception_if_replace_is_true_and_missing_params(
        self, selected_fields, replace_index
    ):
        error_msg = "PostgreSQL ON CONFLICT upsert syntax requires column names and a unique index."
        with pytest.raises(ValueError, match=error_msg):
            _ = BigQueryToPostgresOperator(
                task_id=TASK_ID,
                dataset_table=f"{TEST_DATASET}.{TEST_TABLE_ID}",
                target_table_name=TEST_DESTINATION_TABLE,
                replace=True,
                selected_fields=selected_fields,
                replace_index=replace_index,
            )

    @mock.patch("airflow.providers.google.cloud.hooks.bigquery.BigQueryHook.get_client")
    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.get_credentials_and_project_id"
    )
    @mock.patch("airflow.providers.google.cloud.transfers.bigquery_to_postgres.register_adapter")
    def test_adapters_to_json_registered(self, mock_register_adapter, mock_get_creds, mock_get_client):
        mock_get_creds.return_value = (None, TEST_PROJECT)
        client = MagicMock()
        client.list_rows.return_value = []
        mock_get_client.return_value = client

        operator = BigQueryToPostgresOperator(
            task_id=TASK_ID,
            dataset_table=f"{TEST_DATASET}.{TEST_TABLE_ID}",
            target_table_name=TEST_DESTINATION_TABLE,
            replace=False,
        )
        operator.postgres_hook

        mock_register_adapter.assert_any_call(list, Json)
        mock_register_adapter.assert_any_call(dict, Json)

    @mock.patch("airflow.providers.google.cloud.transfers.bigquery_to_postgres.PostgresHook")
    @mock.patch("airflow.providers.google.cloud.transfers.bigquery_to_sql.BigQueryHook")
    def test_get_openlineage_facets_on_complete_no_selected_fields(self, mock_bq_hook, mock_postgres_hook):
        mock_bq_client = MagicMock()
        mock_bq_client.get_table.return_value = _make_bq_table(["id", "name", "value"])
        mock_bq_hook.get_client.return_value = mock_bq_client
        mock_bq_hook.return_value = mock_bq_hook

        db_info = MagicMock(scheme="postgres", authority="localhost:5432", database="postgresdb")
        mock_postgres_hook.get_openlineage_database_info.return_value = db_info
        mock_postgres_hook.get_openlineage_default_schema.return_value = "postgres-schema"
        mock_postgres_hook.return_value = mock_postgres_hook

        op = BigQueryToPostgresOperator(
            task_id=TASK_ID,
            dataset_table=f"{TEST_DATASET}.{TEST_TABLE_ID}",
            target_table_name="destination",
            selected_fields=None,
            database="postgresdb",
        )
        op.bigquery_hook = mock_bq_hook
        op.bigquery_hook.project_id = TEST_PROJECT
        op.postgres_hook = mock_postgres_hook
        context = mock.MagicMock()
        op.execute(context=context)

        result = op.get_openlineage_facets_on_complete(task_instance=MagicMock())
        assert len(result.inputs) == 1
        assert len(result.outputs) == 1

        input_ds = result.inputs[0]
        assert input_ds.namespace == "bigquery"
        assert input_ds.name == f"{TEST_PROJECT}.{TEST_DATASET}.{TEST_TABLE_ID}"
        assert "schema" in input_ds.facets
        schema_fields = [f.name for f in input_ds.facets["schema"].fields]
        assert set(schema_fields) == {"id", "name", "value"}

        output_ds = result.outputs[0]
        assert output_ds.namespace == "postgres://localhost:5432"
        assert output_ds.name == "postgresdb.postgres-schema.destination"

        assert "columnLineage" in output_ds.facets
        col_lineage = output_ds.facets["columnLineage"]
        assert set(col_lineage.fields.keys()) == {"id", "name", "value"}

    @mock.patch("airflow.providers.google.cloud.transfers.bigquery_to_postgres.PostgresHook")
    @mock.patch("airflow.providers.google.cloud.transfers.bigquery_to_sql.BigQueryHook")
    def test_get_openlineage_facets_on_complete_selected_fields(self, mock_bq_hook, mock_postgres_hook):
        mock_bq_client = MagicMock()
        mock_bq_client.get_table.return_value = _make_bq_table(["id", "name", "value"])
        mock_bq_hook.get_client.return_value = mock_bq_client
        mock_bq_hook.return_value = mock_bq_hook

        db_info = MagicMock(scheme="postgres", authority="localhost:5432", database="postgresdb")
        mock_postgres_hook.get_openlineage_database_info.return_value = db_info
        mock_postgres_hook.get_openlineage_default_schema.return_value = "postgres-schema"
        mock_postgres_hook.return_value = mock_postgres_hook

        op = BigQueryToPostgresOperator(
            task_id=TASK_ID,
            dataset_table=f"{TEST_DATASET}.{TEST_TABLE_ID}",
            target_table_name="destination",
            selected_fields=["id", "name"],
            database="postgresdb",
        )
        op.bigquery_hook = mock_bq_hook
        op.bigquery_hook.project_id = TEST_PROJECT
        op.postgres_hook = mock_postgres_hook
        context = mock.MagicMock()
        op.execute(context=context)

        result = op.get_openlineage_facets_on_complete(task_instance=MagicMock())
        assert len(result.inputs) == 1
        assert len(result.outputs) == 1

        input_ds = result.inputs[0]
        assert input_ds.namespace == "bigquery"
        assert input_ds.name == f"{TEST_PROJECT}.{TEST_DATASET}.{TEST_TABLE_ID}"
        assert "schema" in input_ds.facets
        schema_fields = [f.name for f in input_ds.facets["schema"].fields]
        assert set(schema_fields) == {"id", "name"}

        output_ds = result.outputs[0]
        assert output_ds.namespace == "postgres://localhost:5432"
        assert output_ds.name == "postgresdb.postgres-schema.destination"
        assert "columnLineage" in output_ds.facets
        col_lineage = output_ds.facets["columnLineage"]
        assert set(col_lineage.fields.keys()) == {"id", "name"}
